"""
    Performs basic grid search for ESNs in which the parameter space will be searched in discrete steps.
"""

import numpy as np
import itertools
import operator
import progressbar
from .. import helper as hlp
import matplotlib.pyplot as plt

class GridSearchOptimizer:
    """
    Performs basic grid search for ESNs in which the parameter space will be searched in discrete steps.
    """

    def __init__(self, esnType, parametersDictionary={}, fixedParametersDictionary={}):
        self.esnType = esnType
        self.parametersDictionary = parametersDictionary
        self.fixedParametersDictionary = fixedParametersDictionary


    def fit(self, trainingInput, trainingOutput, validationInput, validationOutput, transientTime, verbose=1):
        """
        Fits an ESN for each of the wanted hyperparameters and predicts the output.
        The best results parameters will be stores in _best_params.
        """

        #calculate the length of all permutations of the hyperparameters
        def enumerate_params():
            keys, values = zip(*self.param_grid.items())
            for row in itertools.product(*values):
                yield dict(zip(keys, row))
        length = sum(1 for x in enumerate_params())

        if verbose > 0:
            #initialize the progressbar to indicate the progress
            bar = progressbar.ProgressBar(max_value=length, redirect_stdout=True)

        #store the results here
        results = []

        for index, params in enumerate(enumerate_params()):
            #create and fit the ESN
            esn = self.esnType(**params, **self.fixed_params)
            trainingAccuricy = esn.fit(trainingInput, trainingOutput, transientTime=transientTime)

            current_state = esn._x

            #evaluate the ESN
            validationMSEs = []

            #check whether only one validation sequence is ought to be checked or if the esn has to be validated on multiple sequences
            if len(validationOutput.shape) == len(trainingInput.shape) + 1:
                for n in range(validationOutput.shape[0]):
                    esn._x = current_state
                    outputPrediction = esn.predict(validationInput[n])
                    validationMSEs.append(np.mean((validationOutput[n] - outputPrediction)**2))
            else:
                 esn._x = current_state
                 outputPrediction = esn.predict(validationInput)
                 validationMSEs.append(np.mean((validationOutput - outputPrediction)**2))

            validationMSE = np.mean(validationMSEs)

            results.append((validationMSE, trainingAccuricy, params))
            
            if verbose > 0:
                bar.update(index)

            #print the currently best result every printfreq step
            if verbose > 1:
                res = min(results, key=operator.itemgetter(0))
                print("Current best parameters: \t: " + str(res))

        if verbose > 0:
            bar.finish()

        #determine the best parameters by minimizing the error
        res = min(results, key=operator.itemgetter(0))

        self._best_params = res[2]
        self._best_mse = res[0]

        return results

    def createDenseHyperparameterGrid(self, parameterGridDictionary):
        """
        Creates a dense grid of hyperparameters for the settings specified inside `parameterGridDictionary`.

        Args:
            parameterGridDictionary (dictionary): Dictionary with strings as keys, which correspond to the name of a hyperparameter of an ESN.
                                                  The values are tuples (start, end, numberOfSteps) which describe the sampling of each parameter's space.
        """
        for key, values in parameterGridDictionary:
            fromParam, tillParam, steps = values
            parameterGridDictionary[key] = np.linspace(fromParam, tillParam, steps)

        self.parametersDictionary = parameterGridDictionary

    @staticmethod
    def plotErrorSurface(esn, N, transientTime, trainInputs, trainTargets, validationInputs, validationTargets, paramDic, gridHeight = None, verbose=1):
        """

        Args:

        Returns:
            trainErrorGrid ():
            validationErrorGrid ():
        """

        if len(list(paramDic.keys())) != 2:
            raise ValueError("Can only plot error surface of exactly 2 hyperparameters")

        if verbose > 0:
            # initialize the progressbar to indicate the progress
            bar = progressbar.ProgressBar(max_value=N*N-1, redirect_stdout=True)

        param1 = list(paramDic.keys())[0]
        fromParam1, tillParam1 = list(paramDic.values())[0]

        param2 = list(paramDic.keys())[1]
        fromParam2, tillParam2 = list(paramDic.values())[1]

        def setParameter(parameter, value):
            if parameter == "spectralRadius":
                esn.setSpectralRadius(value)
            elif parameter == "leakingRate":
                esn.setLeakingRate(value)
            elif parameter == "inputScaling":
                esn.setInputScaling(value)
            else:
                raise ValueError("Hyperparameter {0} does not exist. Choose from either spectralRadius, leakingRate, inputScaling".format( parameter))

        def plotGrid(grid, title, fromParam1, tillParam1, fromParam2, tillParam2, param1, param2, gridHeight=None):
            plotGrid = np.copy(grid)
            if gridHeight is not None:
                maxValue = np.min(grid) + np.min(grid) * gridHeight
                plotGrid[np.where(plotGrid > maxValue)] = maxValue
                plotGrid[np.where(np.isnan(plotGrid))] = maxValue
            fig, ax = plt.subplots()
            mat = plt.imshow(plotGrid, vmin=np.min(plotGrid), vmax=np.max(plotGrid),
                             extent=[fromParam1, tillParam1, tillParam2, fromParam2])
            clb = fig.colorbar(mat)
            plt.title(title)
            plt.xlabel(param1)
            plt.ylabel(param2)
            plt.show()

        trainErrorGrid = np.zeros((N, N))
        validationErrorGrid = np.zeros((N, N))
        widthParam1 = (tillParam1 - fromParam1) / N
        widthParam2 = (tillParam2 - fromParam2) / N
        for i in range(N):
            p1 = i * widthParam1 + fromParam1
            setParameter(param1, p1)
            for j in range(N):
                p2 = j * widthParam2 + fromParam2
                setParameter(param2, p2)

                trainErrorGrid[i, j] = esn.fit(trainInputs, trainTargets, transientTime=transientTime)
                validationErrorGrid[i, j] = hlp.loss(esn.predict(validationInputs), validationTargets)

                if verbose > 0:
                    bar.update(i*N + j)

        if verbose > 0:
            bar.finish()

        plotGrid(trainErrorGrid, "Train error", fromParam1, tillParam1, fromParam2, tillParam2, param1, param2, gridHeight)
        plotGrid(validationErrorGrid, "Validation error", fromParam1, tillParam1, fromParam2, tillParam2, param1, param2, gridHeight)

        return trainErrorGrid, validationErrorGrid



"""
    Implementation of the general ESN model.
"""

import numpy as np
import numpy.random as rnd
from .BaseESN import BaseESN

from . import backend as B

from sklearn.linear_model import Ridge
from sklearn.svm import SVR
from sklearn.linear_model import LogisticRegression
import progressbar
from .OneHotEncoder import OneHotEncoder


class ClassificationESN(BaseESN):
    def __init__(self, n_input, n_reservoir, n_classes,
                 spectralRadius=1.0, noiseLevel=0.0, inputScaling=None,
                 leakingRate=1.0, reservoirDensity=0.2, randomSeed=None,
                 out_activation=lambda x: 0.1+0.98*x/(1+B.exp(-x)), out_inverse_activation=lambda x: B.log((x*0.98+0.01)/(0.99-x*0.98)),
                 weight_generation='naive', bias=1.0, outputBias=1.0,
                 outputInputScaling=1.0, inputDensity=1.0, solver='pinv', regressionParameters={}, activation = B.tanh, activationDerivation=lambda x: 1.0/B.cosh(x)**2):

        super(ClassificationESN, self).__init__(n_input=n_input, n_reservoir=n_reservoir, n_output=n_classes, spectralRadius=spectralRadius,
                                  noiseLevel=noiseLevel, inputScaling=inputScaling, leakingRate=leakingRate, reservoirDensity=reservoirDensity,
                                  randomSeed=randomSeed, out_activation=out_activation, out_inverse_activation=out_inverse_activation,
                                  weight_generation=weight_generation, bias=bias, outputBias=outputBias, outputInputScaling=outputInputScaling,
                                  inputDensity=inputDensity, activation=activation, activationDerivation=activationDerivation)


        self._solver = solver
        self._regressionParameters = regressionParameters
        self._oneHotEncoder = OneHotEncoder()

        """
            allowed values for the solver:
                pinv
                lsqr (will only be used in the thesis)

                sklearn_auto
                sklearn_svd
                sklearn_cholesky
                sklearn_lsqr
                sklearn_sag
        """

    """
        Fits the ESN so that by applying a time series out of inputData the outputData will be produced.

    """
    def fit(self, inputData, outputData, transientTime=0, transientTimeCalculationEpsilon = 1e-3, transientTimeCalculationLength = 20, verbose=0):
        #check the input data
        if inputData.shape[0] != outputData.shape[0]:
            raise ValueError("Amount of input and output datasets is not equal - {0} != {1}".format(inputData.shape[0], outputData.shape[0]))

        nSequences = inputData.shape[0]
        trainingLength = inputData.shape[1]

        if outputData.shape[1] > 1 and outputData.shape[1] != self.n_output:
            raise ValueError("The outputData has the shape {0}, which indicates that it is already one hot encoded, " \
                             "but it does not match the number of classes ({1}) specified in the constructur.".format(outputData.shape, self.n_output))

        if outputData.shape[1] == 1 and len(np.unique(outputData)) > 2:
            outputData = self._oneHotEncoder.transform(outputData)

        self._x = B.zeros((self.n_reservoir,1))

        # Automatic transient time calculations
        if transientTime == "Auto":
            transientTime = self.calculateTransientTime(inputData, outputData, transientTimeCalculationEpsilon,
                                                        transientTimeCalculationLength)
        if transientTime == "AutoReduce":
            if (inputData is None and outputData.shape[1] == 1) or inputData.shape[1] == 1:
                transientTime = self.calculateTransientTime(inputData, outputData, transientTimeCalculationEpsilon,
                                                            transientTimeCalculationLength)
                transientTime = self.reduceTransientTime(inputData, outputData, transientTime)
            else:
                print("Transient time reduction is supported only for 1 dimensional input.")


        self._X = B.zeros((1 + self.n_input + self.n_reservoir, nSequences*(trainingLength-transientTime)))
        Y_target = B.zeros((self.n_output, (trainingLength-transientTime)*nSequences))

        if verbose > 0:
            bar = progressbar.ProgressBar(max_value=len(inputData), redirect_stdout=True, poll_interval=0.0001)
            bar.update(0)

        for n in range(len(inputData)):
            self._x = B.zeros((self.n_reservoir, 1))
            self._X[:, n*(trainingLength-transientTime):(n+1)*(trainingLength-transientTime)] = self.propagate(inputData[n], transientTime, verbose=0)
            #set the target values
            Y_target[:, n*(trainingLength-transientTime):(n+1)*(trainingLength-transientTime)] = np.tile(self.out_inverse_activation(outputData[n]), trainingLength-transientTime).reshape(-1, self.n_output).T

            if verbose > 0:
                bar.update(n)

        if verbose > 0:
            bar.finish()

        if (self._solver == "pinv"):
            self._W_out = B.dot(Y_target, B.pinv(self._X))

            #calculate the training prediction now
            train_prediction = self.out_activation((B.dot(self._W_out, self._X)).T)

        elif (self._solver == "lsqr"):
            X_T = self._X.T
            self._W_out = B.dot(B.dot(Y_target, X_T),B.inv(B.dot(self._X,X_T) + self._regressionParameters[0]*B.identity(1+self.n_input+self.n_reservoir)))

            """
                #alternative represantation of the equation

                Xt = X.T

                A = np.dot(X, Y_target.T)

                B = np.linalg.inv(np.dot(X, Xt)  + regression_parameter*np.identity(1+self.n_input+self.n_reservoir))

                self._W_out = np.dot(B, A)
                self._W_out = self._W_out.T
            """

            #calculate the training prediction now
            train_prediction = self.out_activation(B.dot(self._W_out, self._X).T)

        elif (self._solver in ["sklearn_auto", "sklearn_lsqr", "sklearn_sag", "sklearn_svd"]):
            mode = self._solver[8:]
            params = self._regressionParameters
            params["solver"] = mode
            self._ridgeSolver = Ridge(**params)

            self._ridgeSolver.fit(self._X.T, Y_target.T)

            #calculate the training prediction now
            train_prediction = self.out_activation(self._ridgeSolver.predict(self._X.T))

        elif (self._solver in ["sklearn_svr", "sklearn_svc"]):
            self._ridgeSolver = SVR(**self._regressionParameters)

            self._ridgeSolver.fit(self._X.T, Y_target.T.ravel())

            #calculate the training prediction now
            train_prediction = self.out_activation(self._ridgeSolver.predict(self._X.T))

        train_prediction = train_prediction[::trainingLength]

        #calculate the training error now
        training_error = B.sqrt(B.mean((train_prediction - outputData)**2))
        return training_error


    """
        Use the ESN in the predictive mode to predict the output signal by using an input signal.
    """
    def predict(self, inputData, update_processor=lambda x:x, transientTime=0, verbose=0):
        if (len(inputData.shape) == 1):
            inputData = inputData[None, :]

        predictionLength = inputData.shape[1]

        Y = B.empty((inputData.shape[0], self.n_output))

        if (verbose > 0):
            bar = progressbar.ProgressBar(max_value=inputData.shape[0], redirect_stdout=True, poll_interval=0.0001)
            bar.update(0)

        for n in range(inputData.shape[0]):
            #reset the state
            self._x = B.zeros(self._x.shape)

            X = self.propagate(inputData[n], transientTime)
            #calculate the prediction using the trained model
            if (self._solver in ["sklearn_auto", "sklearn_lsqr", "sklearn_sag", "sklearn_svd", "sklearn_svr"]):
                y = self._ridgeSolver.predict(X.T).reshape((self.n_output, -1))
            else:
                y = B.dot(self._W_out, X)

            Y[n] = np.mean(y, 1)

            if verbose > 0:
                bar.update(n)

        if verbose > 0:
            bar.finish()

        #return the result
        result = B.zeros(Y.shape)
        for i,n in enumerate(B.argmax(Y, 1)): 
            result[i, n] = 1.0
        return result



"""
    Implementation of the general ESN model.
"""

import numpy as np
import numpy.random as rnd
from .BaseESN import BaseESN

from . import backend as B

from sklearn.linear_model import Ridge
from sklearn.svm import SVR
from sklearn.linear_model import LogisticRegression
import progressbar

#import dill

from multiprocess import Process, Queue, Manager, Pool, cpu_count #we require Pathos version >=0.2.6. Otherwise we will get an "EOFError: Ran out of input" exception
#import multiprocessing
import ctypes
from multiprocessing import process

class SpatioTemporalESN(BaseESN):
    def __init__(self, inputShape, n_reservoir,
                 filterSize=1, stride=1, borderMode="mirror", nWorkers="auto",
                 spectralRadius=1.0, noiseLevel=0.0, inputScaling=None,
                 leakingRate=1.0, reservoirDensity=0.2, randomSeed=None, averageOutputWeights=True,
                 out_activation=lambda x: x, out_inverse_activation=lambda x: x,
                 weightGeneration='naive', bias=1.0, outputBias=1.0,
                 outputInputScaling=1.0, input_density=1.0, solver='pinv', regressionParameters={}, activation = B.tanh, activationDerivation=lambda x: 1.0/B.cosh(x)**2):
       
        self._averageOutputWeights = averageOutputWeights
        if averageOutputWeights and solver != "lsqr":
            raise ValueError("`averageOutputWeights` can only be set to `True` when `solver` is set to `lsqr` (Ridge Regression)")

        self._borderMode = borderMode
        if not borderMode in ["mirror", "padding", "edge", "wrap"]:
            raise ValueError("`borderMode` must be set to one of the following values: `mirror`, `padding`, `edge` or `wrap`.")

        self._regressionParameters = regressionParameters
        self._solver = solver

        n_inputDimensions = len(inputShape)

        if filterSize % 2 == 0:
            raise ValueError("filterSize has to be an odd number (1, 3, 5, ...).")
        self._filterSize = filterSize
        self._filterWidth = int(np.floor(filterSize / 2))
        self._stride = stride

        self._n_input = int(np.power(np.ceil(filterSize / stride), n_inputDimensions))

        self.n_inputDimensions = n_inputDimensions
        self.inputShape = inputShape

        if not self._averageOutputWeights:
            self._WOuts = B.empty((np.prod(inputShape), 1, self._n_input+n_reservoir+1))
            self._WOut = None
        else:
            self._WOuts = None
            self._WOut = B.zeros((1, self._n_input+n_reservoir+1))
        self._xs = B.empty((np.prod(inputShape), n_reservoir, 1))

        
        if nWorkers=="auto":
            self._nWorkers = np.max((cpu_count()-1, 1))
        else:
            self._nWorkers = nWorkers
            

        manager = Manager()
        self.sharedNamespace = manager.Namespace()
        if hasattr(self, "fitWorkerID") == False or self.parallelWorkerIDs is None:
            self.parallelWorkerIDs = manager.Queue()
            for i in range(self._nWorkers):
                self.parallelWorkerIDs.put((i))

        super(SpatioTemporalESN, self).__init__(n_input=self._n_input, n_reservoir=n_reservoir, n_output=1, spectralRadius=spectralRadius,
                                  noiseLevel=noiseLevel, inputScaling=inputScaling, leakingRate=leakingRate, reservoirDensity=reservoirDensity,
                                  randomSeed=randomSeed, out_activation=out_activation, out_inverse_activation=out_inverse_activation,
                                  weightGeneration=weightGeneration, bias=bias, outputBias=outputBias, outputInputScaling=outputInputScaling,
                                  input_density=input_density, activation=activation, activationDerivation=activationDerivation)

        """
            allowed values for the solver:
                pinv
                lsqr (will only be used in the thesis)

                sklearn_auto
                sklearn_svd
                sklearn_cholesky
                sklearn_lsqr
                sklearn_sag
        """  
    
    def resetState(self, index=None):
        if index is None:
             self._x = B.zeros((self._nWorkers, self.n_reservoir, 1))
        else:
            self._x[index] = B.zeros((self.n_reservoir, 1))

    def _embedInputData(self, inputData):
        rank = len(inputData.shape) - 1

        if self._borderMode == "mirror":
            modifiedInputData = np.pad(inputData, tuple([(0,0)] + [(self._filterWidth, self._filterWidth)]*rank), mode="symmetric")
        elif self._borderMode == "padding":
            modifiedInputData = np.pad(inputData, tuple([(0,0)] + [(self._filterWidth, self._filterWidth)]*rank), mode="constant", constant_values=0)
        elif self._borderMode == "edge":
            modifiedInputData = np.pad(inputData, tuple([(0,0)] + [(self._filterWidth, self._filterWidth)]*rank), mode="edge")
        elif self._borderMode == "wrap":
            modifiedInputData = np.pad(inputData, tuple([(0,0)] + [(self._filterWidth, self._filterWidth)]*rank), mode="wrap")

        return modifiedInputData

    """
        Fits the ESN so that by applying a time series out of inputData the outputData will be produced.
    """
    def fit(self, inputData, outputData, transientTime=0, verbose=0):
        rank = len(inputData.shape) - 1

        if rank != self.n_inputDimensions:
            raise ValueError("The `inputData` does not have a suitable shape. It has to have {0} spatial dimensions and 1 temporal dimension.".format(self.n_inputDimensions))

        manager = Manager()
        fitQueue = manager.Queue()

        modifiedInputData = self._embedInputData(inputData)
        
        self.sharedNamespace.inputData = modifiedInputData
        self.sharedNamespace.outputData = outputData
        self.sharedNamespace.transientTime = transientTime

        self.sharedNamespace.WOut = self._WOut
        self.sharedNamespace.WOuts = self._WOuts
        self.sharedNamespace.xs = self._xs
               
        jobs = np.stack(np.meshgrid(*[np.arange(x)+self._filterWidth for x in inputData.shape[1:]]), axis=rank).reshape(-1, rank).tolist()
        nJobs = len(jobs)

        self.sharedNamespace.WOuts = self._WOuts

        self.resetState()
        
        pool = Pool(processes=self._nWorkers, initializer=SpatioTemporalESN._init_fitProcess, initargs=[fitQueue, self])  
        pool.map_async(self._fitProcess, jobs)

        def _processPoolWorkerResults():
            nJobsDone = 0
        
            if verbose > 0:
                bar = progressbar.ProgressBar(max_value=nJobs, redirect_stdout=True, poll_interval=0.0001)
                bar.update(0)

            while nJobsDone < nJobs:
                data = fitQueue.get()
            
                #result of fitting
                indices, x, WOut = data
                id = self._uniqueIDFromIndices(indices)
                
                #store WOut
                if self._averageOutputWeights:
                    self._WOut += WOut/np.prod(self.inputShape)
                else:
                    self._WOuts[id] = WOut 

                #store x
                self._xs[id] = x
           
                nJobsDone += 1
                if verbose > 0:
                    bar.update(nJobsDone)
                    if verbose > 1:
                        print(nJobsDone)

            if verbose > 0:
                bar.finish()
        _processPoolWorkerResults()

        pool.close()

        self.sharedNamespace.WOut = self._WOut
        self.sharedNamespace.WOuts = self._WOuts
        self.sharedNamespace.xs = self._xs

    """
        Use the ESN in the predictive mode to predict the output signal by using an input signal.
    """
    def predict(self, inputData, transientTime=0, update_processor=lambda x:x, verbose=0):
        rank = len(inputData.shape) - 1

        if rank != self.n_inputDimensions:
            raise ValueError("The `inputData` does not have a suitable shape. It has to have {0} spatial dimensions and 1 temporal dimension.".format(self.n_inputDimensions))

        manager = Manager()
        predictQueue = manager.Queue()

        modifiedInputData = self._embedInputData(inputData)

        self.sharedNamespace.inputData = modifiedInputData
        self.sharedNamespace.transientTime = transientTime
        predictionOutput = B.zeros(np.insert(self.inputShape, 0, inputData.shape[0]-transientTime))
       
        jobs = np.stack(np.meshgrid(*[np.arange(x)+self._filterWidth for x in inputData.shape[1:]]), axis=rank).reshape(-1, rank).tolist()
        nJobs = len(jobs)

        self.resetState()

        pool = Pool(processes=self._nWorkers, initializer=SpatioTemporalESN._init_predictProcess, initargs=[predictQueue, self])  
        pool.map_async(self._predictProcess, jobs)
        
        def _processPoolWorkerResults():
            nJobsDone = 0
        
            if verbose > 0:
                bar = progressbar.ProgressBar(max_value=nJobs, redirect_stdout=True, poll_interval=0.0001)
                bar.update(0)

            while nJobsDone < nJobs:
                data = predictQueue.get()
                #result of predicting
                indices, prediction = data
                #update the values
                predictionOutput[tuple([Ellipsis] +  indices)] = prediction
                         
                nJobsDone += 1
                if verbose > 0:
                    bar.update(nJobsDone)
                    if verbose > 1:
                        print(nJobsDone)
        
            if verbose > 0:
                bar.finish()   
        _processPoolWorkerResults()

        pool.close()

        return predictionOutput
    
    def _uniqueIDFromIndices(self, indices):
        id = indices[-1]

        if len(indices) != len(self.inputShape):
            raise ValueError("Shape if `indices` does not match the `inputShape` of the SpatioTemporalESN.")

        if len(self.inputShape) > 1:
            for i in range(0, len(self.inputShape)-1):
                id += self.inputShape[i+1]*indices[i]

        return id

    @staticmethod
    def _init_fitProcess(fitQueue, self):
        SpatioTemporalESN._fitProcess.fitQueue = fitQueue
        SpatioTemporalESN._fitProcess.self = self

    def _fitProcess(self, indices):
        inputData = self.sharedNamespace.inputData
        outputData = self.sharedNamespace.outputData
        transientTime = self.sharedNamespace.transientTime
       
        y, x = indices

        #print(id(SpatioTemporalESN._fitProcess.sharedNamespace.parallelWorkerIDs))
        workerID = self.parallelWorkerIDs.get()
        #self.sharedNamespace.parallelWorkerIDs = self.sharedNamespace.parallelWorkerIDs
       
        #create patchedInputData

        #treat the frame pixels in a special way
        inData = inputData[:, y-self._filterWidth : y+self._filterWidth+1, x-self._filterWidth : x+self._filterWidth+1][:, ::self._stride, ::self._stride].reshape(len(inputData), -1)
        #create target output series
        outData = outputData[:, y-self._filterWidth, x-self._filterWidth].reshape(-1, 1)
       
        #now fit
        X = self.propagate(inData, transientTime=transientTime, x=self._x[workerID], verbose=0)

        #define the target values
        Y_target = self.out_inverse_activation(outData).T[:, transientTime:]

        X_T = X.T

        WOut = B.dot(B.dot(Y_target, X_T),B.inv(B.dot(X, X_T) + self._regressionParameters[0]*B.identity(1+self.n_input+self.n_reservoir)))
        
        #calculate the training prediction now
        trainingPrediction = self.out_activation(B.dot(WOut, X).T)
       
        #store the state and the output matrix of the worker
        SpatioTemporalESN._fitProcess.fitQueue.put(([x-self._filterWidth for x in indices], self._x[workerID].copy(), WOut.copy()))
     
        self.parallelWorkerIDs.put(workerID)

    @staticmethod
    def _init_predictProcess(predictQueue, self):
        SpatioTemporalESN._predictProcess.predictQueue = predictQueue
        SpatioTemporalESN._predictProcess.self = self

    def _predictProcess(self, indices):
        inputData = self.sharedNamespace.inputData
        transientTime = self.sharedNamespace.transientTime

        y, x = indices
        workerID = self.parallelWorkerIDs.get()
        #get internal id
 
        id = self._uniqueIDFromIndices([x-self._filterWidth for x in indices])
               
        #create patchedInputData
        #treat the frame pixels in a special way
        inData = inputData[:, y-self._filterWidth:y+self._filterWidth+1, x-self._filterWidth:x+self._filterWidth+1][:, ::self._stride, ::self._stride].reshape(len(inputData), -1)

        self._x[workerID] = self.sharedNamespace.xs[id]
       
        #now fit
        X = self.propagate(inData, transientTime, x=self._x[workerID], verbose=0)
        self.sharedNamespace.xs[id] = self._x[workerID]
        
        if self._averageOutputWeights:
            WOut = self._WOut
        else:
            WOut = self._WOuts[id]

        #calculate the actual prediction
        prediction = self.out_activation(B.dot(WOut, X).T)[:, 0]
            
        #store the state and the output matrix of the worker
        SpatioTemporalESN._predictProcess.predictQueue.put(([x-self._filterWidth for x in indices], prediction))
        
        self.parallelWorkerIDs.put(workerID)


import numpy as np
from .. import helper as hlp
from .. import backend as B

import progressbar

class GradientOptimizer(object):
    def _validateReservoir(self):
        if not np.isclose(self._reservoir._noiseLevel, 0.0):
            raise ValueError("Noise must be set to 0.0 for this optimizer, but it is set to {0}.".format(self._reservoir._noiseLevel))
        if self._reservoir._solver is not "lsqr":
            raise ValueError("The reservoir's solver must be set to 'lsqr' (Ridge Regression) for this optimizer.")

    def __init__(self, reservoir, learningRate=0.0001):
        self._reservoir = reservoir
        self._validateReservoir()
        self.setLearningRate(learningRate)


    def setLearningRate(self, learningRate):
        if np.isscalar(learningRate):
            self._learningRates = (learningRate, learningRate, learningRate)
        else:
            if len(learningRate) != 3:
                raise ValueError("LearningRate has to be a scalar or a list/tuple with 3 entries.")
            else:
                self._learningRates = learningRate

    ####################################################################################################################################################

    def _derivationLrSrIsFs(self, oldDerivatives, W_uniform, W_in_uniform, u, x, generative=False, W_fb_uniform = 0, oldy = 0):
        doldLr, doldSr, doldIs, doldFs = oldDerivatives
        a = self._reservoir._leakingRate
        X = self._reservoir.calculateLinearNetworkTransmissions(u)

        activationX = self._reservoir._activation(X)
        activationDerivationX = self._reservoir._activationDerivation(X)

        dLr = (1-a) * doldLr - x + activationX + a * activationDerivationX * B.dot(self._reservoir._W, doldLr)
        dSr = (1-a) * doldSr + a * activationDerivationX * (B.dot(self._reservoir._W, doldSr) + B.dot(W_uniform, x))

        u = B.vstack((1,u))
        dIs = (1-a) * doldIs + a * activationDerivationX * (B.dot(self._reservoir._W, doldIs) + B.dot(W_in_uniform, u))

        if generative:
            dFs = (1 - a) * doldFs + a * activationDerivationX * ( B.dot(self._reservoir.W, doldFs) + B.dot(W_fb_uniform, oldy) )
            return dLr, dSr, dIs, dFs

        return dLr, dSr, dIs

    # del W_out / del beta
    def derivationForPenalty(self, Y, X, penalty):
        X_T = X.T
        term2 = B.inv( B.dot( X, X_T ) + penalty * B.identity(1 + self._reservoir.n_input + self._reservoir.n_self._reservoir) )
        return - B.dot( B.dot( Y, X_T ), B.dot( term2, term2 ) )

    # del W_out / del (alpha, rho or s_in)
    def derivationWoutForP(self, Y, X, XPrime):
        X_T = X.T
        XPrime_T = XPrime.T

        # A = dot(X,X_T) + penalty*eye(1 + self.target_dim + self.n_reservoir)
        # APrime = dot( XPrime, X_T) + dot( X, XPrime_T )
        # APrime_T = APrime.T
        # InvA = B.inv(A)
        # InvA_T = InvA.T
        #
        # term1 = dot(XPrime_T, InvA)
        #
        # term21 = -dot( InvA, dot( APrime, InvA ) )
        # term22 = dot( dot( dot( InvA, InvA_T), APrime_T), eye(1 + self.target_dim + self.n_reservoir) - dot( A, InvA ) )
        # term23 = dot( dot( eye(1 + self.target_dim + self.n_reservoir) - dot( InvA, A ), APrime_T), dot( InvA_T, InvA) )
        # term2 = dot( X_T, term21 + term22 + term23 )
        #
        # return dot( Y, term1 + term2)

        term1 = B.inv(B.dot(X,X_T) + self._reservoir._regressionParameters[0]*B.identity(1 + self._reservoir.n_input + self._reservoir.n_reservoir))
        term2 = B.dot( XPrime, X_T) + B.dot( X, XPrime_T )

        return B.dot( Y, B.dot( XPrime_T, term1 ) - B.dot( B.dot( B.dot( X_T, term1 ), term2 ), term1 ) )
    ####################################################################################################################################################


    def fit(self, trainingInput, trainingOutput, validationInput, validationOutput, verbose=1):
        self.optimizeParameterForTrainError(trainingInput, trainingOutput, validationInput, validationOutput, verbose)
        self.optimizePenaltyForEvaluationError(trainingInput, trainingOutput, validationInput, validationOutput, verbose)
        self.optimizeParameterForValidationError(trainingInput, trainingOutput, validationInput, validationOutput, verbose)

    def optimizeParameterForTrainError(self, trainingInputData, trainingOutputData, validationInputData, validationOutputData,
                                       epochs=1, transientTime=None, verbose=1):

        self._validateReservoir()

        learningRate = self._learningRates[0]

        if not np.isscalar(self._reservoir._inputScaling):
            raise ValueError("Only penalty optimization is supported for a multiple input scalings at the moment. We are working on it.")

        # calculate stuff
        trainLength = trainingInputData.shape[0]
        if transientTime is None:
            transientTime = self._reservoir.estimateTransientTime(trainingInputData, trainingOutputData)

        # initializations of arrays:
        if (len(trainingOutputData.shape) == 1):
            Ytarget = trainingOutputData[None, transientTime:].T
        else:
            Ytarget = trainingOutputData[transientTime:].T

        if (len(validationOutputData.shape) == 1):
            validationOutputData = validationOutputData[:, None]

        Ytarget = self._reservoir.out_inverse_activation(Ytarget)
        validationOutputData = self._reservoir.out_inverse_activation(validationOutputData)

        originalLearningRate = learningRate

        # initializations for plotting parameter and losses at the end
        inputScalings = list()
        leakingRates = list()
        spectralRadiuses = list()
        fitLosses = list()
        validationLosses = list()
        learningRates = list()

        # initializations for arrays which collect all the gradients of the error of the single time steps, which get add at the end
        srGradients = B.zeros(trainLength - transientTime)
        lrGradients = B.zeros(trainLength - transientTime)
        isGradients = B.zeros(trainLength - transientTime)

        # collecting the single derivatives  - > this is the derivation of design matrix when filled
        srGradientsMatrix = B.zeros((self._reservoir.n_reservoir + self._reservoir.n_input + 1, trainLength - transientTime))
        lrGradientsMatrix = B.zeros((self._reservoir.n_reservoir + self._reservoir.n_input + 1, trainLength - transientTime))
        isGradientsMatrix = B.zeros((self._reservoir.n_reservoir + self._reservoir.n_input + 1, trainLength - transientTime))

        # initialize fallback Parameter
        oldSR = self._reservoir._spectralRadius
        oldLR = self._reservoir._leakingRate
        oldIS = self._reservoir._inputScaling

        # initialize self.designMatrix and self.W_out
        oldLoss = self._reservoir.fit(trainingInputData, trainingOutputData, transientTime=transientTime)

        # Calculate uniform matrices
        W_uniform = self._reservoir._W / self._reservoir._spectralRadius
        W_in_uniform = self._reservoir._WInput / self._reservoir._inputScaling

        if (verbose > 0):
            bar = progressbar.ProgressBar(max_value=epochs, redirect_stdout=True, poll_interval=0.0001)
            bar.update(0)

        for epoch in range(epochs):
            if verbose > 0:
                bar.update(epoch)

            # initialize del x / del (rho, alpha, s_in) and reservoir state itself
            derivationSpectralRadius = B.zeros((self._reservoir.n_reservoir, 1))
            derivationLeakingRate = B.zeros((self._reservoir.n_reservoir, 1))
            derivationInputScaling = B.zeros((self._reservoir.n_reservoir, 1))

            # initialize the neuron states new
            x = B.zeros((self._reservoir.n_reservoir, 1))

            # go thorugh the train length (e.g. the time, on which W_out gets calculated)
            for t, u in enumerate(trainingInputData):
                u = u.reshape(-1, 1)
                oldx = x

                #apply the update method by using the input and the output
                self._reservoir.update(u, trainingOutputData[t])
                x = self._reservoir._x
                # x = self._reservoir._X[2:,t]


                # calculate the del /x del (rho, alpha, s_in)
                #derivationSpectralRadius = self.derivationForSpectralRadius(W_uniform, derivationSpectralRadius, u, oldx)
                #derivationLeakingRate = self.derivationForLeakingRate(derivationLeakingRate, u, oldx)
                #derivationInputScaling = self.derivationForInputScaling(W_in_uniform, derivationInputScaling, u, oldx)

                derivationLeakingRate, derivationSpectralRadius, derivationInputScaling = self._derivationLrSrIsFs( (derivationLeakingRate, derivationSpectralRadius, derivationInputScaling, None), W_uniform, W_in_uniform, u, oldx)
                if t >= transientTime:

                    # concatenate with zeros (for the derivatives of the input and the 1, which are always 0)
                    derivationConcatinationSpectralRadius = B.concatenate( (B.zeros(self._reservoir.n_input + 1), derivationSpectralRadius[:, 0]), axis=0)
                    derivationConcatinationLeakingRate = B.concatenate( (B.zeros(self._reservoir.n_input + 1), derivationLeakingRate[:, 0]), axis=0)
                    derivationConcatinationInputScaling = B.concatenate( (B.zeros(self._reservoir.n_input + 1), derivationInputScaling[:, 0]), axis=0)

                    # add to matrix
                    srGradientsMatrix[:, t - transientTime] = derivationConcatinationSpectralRadius
                    lrGradientsMatrix[:, t - transientTime] = derivationConcatinationLeakingRate
                    isGradientsMatrix[:, t - transientTime] = derivationConcatinationInputScaling

            # calculate del W_out / del (rho, alpha, s_in) based on the designMatrix and the derivative of the designMatrix we just calculated
            WoutPrimeSR = self.derivationWoutForP(Ytarget, self._reservoir._X, srGradientsMatrix)
            WoutPrimeLR = self.derivationWoutForP(Ytarget, self._reservoir._X, lrGradientsMatrix)
            WoutPrimeIS = self.derivationWoutForP(Ytarget, self._reservoir._X, isGradientsMatrix)

            # reinitialize the states
            x = B.zeros((self._reservoir.n_reservoir, 1))

            # go through the train time again, and this time, calculate del error / del (rho, alpha, s_in) based on del W_out and the single derivatives
            for t, u in enumerate(trainingInputData):
                u = u.reshape(-1, 1)
                self._reservoir.update(u, trainingOutputData[t])
                x = self._reservoir._x
                if t >= transientTime:

                    # calculate error at given time step
                    error = (trainingOutputData[t] - B.dot( self._reservoir._WOut, B.vstack((1, u, x)) ) ).T

                    # calculate gradients
                    gradientSR = B.dot(-error, B.dot(WoutPrimeSR, B.vstack((1, u, x))[:, 0]) + B.dot(self._reservoir._WOut, srGradientsMatrix[:, t - transientTime]))
                    srGradients[t - transientTime] = gradientSR
                    gradientLR = B.dot(-error, B.dot(WoutPrimeLR, B.vstack((1, u, x))[:, 0]) + B.dot(self._reservoir._WOut, lrGradientsMatrix[:, t - transientTime]))
                    lrGradients[t - transientTime] = gradientLR
                    gradientIS = B.dot(-error, B.dot(WoutPrimeIS, B.vstack((1, u, x))[:, 0]) + B.dot(self._reservoir._WOut, isGradientsMatrix[:, t - transientTime]))
                    isGradients[t - transientTime] = gradientIS

            # sum up the gradients del error / del (rho, alpha, s_in) to get final gradient
            gradientSR = sum(srGradients)
            gradientLR = sum(lrGradients)
            gradientIS = sum(isGradients)

            # normalize gradients to length 1
            gradientVectorLength = B.sqrt(gradientSR ** 2 + gradientLR ** 2 + gradientIS ** 2)

            gradientSR /= gradientVectorLength
            gradientLR /= gradientVectorLength
            gradientIS /= gradientVectorLength

            # update spectral radius
            self._reservoir.setSpectralRadius(self._reservoir._spectralRadius - learningRate * gradientSR)

            # update leaking rate
            self._reservoir.setLeakingRate( self._reservoir._leakingRate - learningRate * gradientLR )

            # update input scaling
            self._reservoir.setInputScaling(self._reservoir._inputScaling - learningRate * gradientIS)

            # calculate the errors and update the self.designMatrix and the W_out
            fitLoss = self._reservoir.fit(trainingInputData, trainingOutputData, transientTime=transientTime)
            validationLoss = hlp.loss( self._reservoir.predict(validationInputData), validationOutputData )

            if fitLoss > oldLoss:
                self._reservoir.setSpectralRadius(oldSR)
                self._reservoir.setLeakingRate(oldLR)
                self._reservoir.setInputScaling(oldIS)
                learningRate = learningRate / 2
            else:
                learningRate = min(learningRate*1.5, originalLearningRate)
               
                oldSR = self._reservoir._spectralRadius
                oldLR = self._reservoir._leakingRate
                oldIS = self._reservoir._inputScaling
                oldLoss = fitLoss
                spectralRadiuses.append(self._reservoir._spectralRadius)
                leakingRates.append(self._reservoir._leakingRate)
                inputScalings.append(self._reservoir._inputScaling)
                fitLosses.append(fitLoss)
                validationLosses.append(validationLoss)
            learningRates.append(learningRate)

        if verbose > 0:
            bar.finish()

        return (validationLosses, fitLosses, inputScalings, leakingRates, spectralRadiuses, learningRates)

    def optimizeParameterForValidationError(self, trainingInputData, trainingOutputData, validationInputData, validationOutputData,
                                            optimizationLength, epochs=1, transientTime=None, verbose=1):

        self._validateReservoir()

        learningRate = self._learningRates[1]

        if not np.isscalar(self._reservoir._inputScaling):
            raise ValueError("Only penalty optimization is supported for a multiple input scalings at the moment. We are working on it.")

        # calculate stuff
        trainLength = trainingInputData.shape[0]
        if transientTime is None:
            transientTime = self._reservoir.estimateTransientTime(trainingInputData, trainingOutputData)

        # initializations of arrays:
        if (len(trainingOutputData.shape) == 1):
            Ytarget = trainingOutputData[None, transientTime:].T
        else:
            Ytarget = trainingOutputData[transientTime:].T

        if (len(validationOutputData.shape) == 1):
            validationOutputData = validationOutputData[:, None]

        Ytarget = self._reservoir.out_inverse_activation(Ytarget)
        validationOutputData = self._reservoir.out_inverse_activation(validationOutputData)

        originalLearningRate = learningRate

        # initializations for plotting parameter and losses at the end
        inputScalings = list()
        leakingRates = list()
        spectralRadiuses = list()
        fitLosses = list()
        validationLosses = list()
        learingRates = list()

        # initializations for arrays which collect all the gradients of the error of the single time steps, which get add at the end
        srGradients = B.zeros(optimizationLength)
        lrGradients = B.zeros(optimizationLength)
        isGradients = B.zeros(optimizationLength)

        # collecting the single derivatives  - > this is the derivation of design matrix when filled
        srGradientsMatrix = B.zeros((self._reservoir.n_reservoir + self._reservoir.n_input + 1, trainLength - transientTime))
        lrGradientsMatrix = B.zeros((self._reservoir.n_reservoir + self._reservoir.n_input + 1, trainLength - transientTime))
        isGradientsMatrix = B.zeros((self._reservoir.n_reservoir + self._reservoir.n_input + 1, trainLength - transientTime))

        # initialize variables for the "when the error goes up, go back and divide learning rate by 2" mechanism
        oldSR = self._reservoir._spectralRadius
        oldLR = self._reservoir._leakingRate
        oldIS = self._reservoir._inputScaling

        # initialize self.designMatrix and self.W_out
        self._reservoir.fit(trainingInputData, trainingOutputData, transientTime=transientTime)
        oldLoss = hlp.loss( self._reservoir.predict(validationInputData), validationOutputData )

        # Calculate uniform matrices
        W_uniform = self._reservoir._W / self._reservoir._spectralRadius
        W_in_uniform = self._reservoir._WInput / self._reservoir._inputScaling

        if (verbose > 0):
            bar = progressbar.ProgressBar(max_value=epochs, redirect_stdout=True, poll_interval=0.0001)
            bar.update(0)

        for epoch in range(epochs):
            if verbose > 0:
                bar.update(epoch)

            # initialize del x / del (rho, alpha, s_in) and reservoir state itself
            derivationSpectralRadius = B.zeros((self._reservoir.n_reservoir, 1))
            derivationLeakingRate = B.zeros((self._reservoir.n_reservoir, 1))
            derivationInputScaling = B.zeros((self._reservoir.n_reservoir, 1))
            x = B.zeros((self._reservoir.n_reservoir, 1))

            # go thorugh the train length (e.g. the time, on which W_out gets calculated)
            for t, u in enumerate(trainingInputData):
                u = u.reshape(-1, 1)
                oldx = x
                x = self._reservoir._X[2:,t]


                # calculate the del /x del (rho, alpha, s_in)
                #derivationSpectralRadius = self.derivationForSpectralRadius(W_uniform, derivationSpectralRadius, u, oldx)
                #derivationLeakingRate = self.derivationForLeakingRate(derivationLeakingRate, u, oldx)
                #derivationInputScaling = self.derivationForInputScaling(W_in_uniform, derivationInputScaling, u, oldx)

                derivationLeakingRate, derivationSpectralRadius, derivationInputScaling = self._derivationLrSrIsFs( (derivationLeakingRate, derivationSpectralRadius, derivationInputScaling, None), W_uniform, W_in_uniform, u, oldx)
                if t >= transientTime:

                    # concatenate with zeros (for the derivatives of the input and the 1, which are always 0)
                    srGradientsMatrix[:, t - transientTime] = B.concatenate( (B.zeros(self._reservoir.n_input + 1), derivationSpectralRadius[:, 0]), axis=0)
                    lrGradientsMatrix[:, t - transientTime] = B.concatenate( (B.zeros(self._reservoir.n_input + 1), derivationLeakingRate[:, 0]), axis=0)
                    isGradientsMatrix[:, t - transientTime] = B.concatenate( (B.zeros(self._reservoir.n_input + 1), derivationInputScaling[:, 0]), axis=0)

            # add to matrix
            WoutPrimeSR = self.derivationWoutForP(Ytarget, self._reservoir._X, srGradientsMatrix)
            WoutPrimeLR = self.derivationWoutForP(Ytarget, self._reservoir._X, lrGradientsMatrix)
            WoutPrimeIS = self.derivationWoutForP(Ytarget, self._reservoir._X, isGradientsMatrix)

            # this time go through validation length
            for t, u in enumerate(validationInputData):
                u = u.reshape(-1, 1)
                oldx = x
                self._reservoir.update(u)
                x = self._reservoir._x

                # calculate error at given time step
                error = (validationOutputData[t] - B.dot( self._reservoir._WOut, B.vstack((1, u, x)) ) ).T

# calculate the del /x del (rho, alpha, s_in)
                #derivationSpectralRadius = self.derivationForSpectralRadius(W_uniform, derivationSpectralRadius, u, oldx)
                #derivationLeakingRate = self.derivationForLeakingRate(derivationLeakingRate, u, oldx)
                #derivationInputScaling = self.derivationForInputScaling(W_in_uniform, derivationInputScaling, u, oldx)

                derivationLeakingRate, derivationSpectralRadius, derivationInputScaling = self._derivationLrSrIsFs((derivationLeakingRate, derivationSpectralRadius, derivationInputScaling, None), W_uniform, W_in_uniform, u, oldx)

                # concatenate derivations with 0
                derivationConcatinationSpectralRadius = B.concatenate( (B.zeros(self._reservoir.n_input + 1), derivationSpectralRadius[:, 0]), axis=0)
                derivationConcatinationLeakingRate = B.concatenate( (B.zeros(self._reservoir.n_input + 1), derivationLeakingRate[:, 0]), axis=0)
                derivationConcatinationInputScaling = B.concatenate( (B.zeros(self._reservoir.n_input + 1), derivationInputScaling[:, 0]), axis=0)

                # calculate gradients
                gradientSR = B.dot(-error, B.dot(self._reservoir._WOut, derivationConcatinationSpectralRadius) + B.dot(WoutPrimeSR, B.vstack((1, u, x))[:, 0]))
                srGradients[t] = gradientSR
                gradientLR = B.dot(-error, B.dot(self._reservoir._WOut, derivationConcatinationLeakingRate) + B.dot(WoutPrimeLR, B.vstack( (1, u, x))[:, 0]))
                lrGradients[t] = gradientLR
                gradientIS = B.dot(-error, B.dot(self._reservoir._WOut, derivationConcatinationInputScaling) + B.dot(WoutPrimeIS, B.vstack( (1, u, x))[:, 0]))
                isGradients[t] = gradientIS

            # sum up the gradients del error / del (rho, alpha, s_in) to get final gradient
            gradientSR = sum(srGradients)
            gradientLR = sum(lrGradients)
            gradientIS = sum(isGradients)

            # normalize length of gradient to 1
            gradientVectorLength = B.sqrt(gradientSR ** 2 + gradientLR ** 2 + gradientIS ** 2)

            gradientSR /= gradientVectorLength
            gradientLR /= gradientVectorLength
            gradientIS /= gradientVectorLength

            # update spectral radius
            self._reservoir.setSpectralRadius(self._reservoir._spectralRadius - learningRate * gradientSR)

            # update leaking rate
            self._reservoir.setLeakingRate(self._reservoir._leakingRate - learningRate * gradientLR)

            # update input scaling
            self._reservoir.setInputScaling(self._reservoir._inputScaling - learningRate * gradientIS)

            # calculate the errors and update the self.designMatrix and the W_out
            fitLoss = self._reservoir.fit(trainingInputData, trainingOutputData, transientTime=transientTime)
            validationLoss = hlp.loss( self._reservoir.predict(validationInputData), validationOutputData )

            # this is the "when the error goes up, go back and divide learning rate by 2" mechanism
            if validationLoss > oldLoss:
                self._reservoir.setSpectralRadius(oldSR)
                self._reservoir.setLeakingRate(oldLR)
                self._reservoir.setInputScaling(oldIS)
                learningRate = learningRate / 2
            else:
                learningRate = min(learningRate*1.5, originalLearningRate)

                oldSR = self._reservoir._spectralRadius
                oldLR = self._reservoir._leakingRate
                oldIS = self._reservoir._inputScaling
                oldLoss = validationLoss

                spectralRadiuses.append(self._reservoir._spectralRadius)
                leakingRates.append(self._reservoir._leakingRate)
                inputScalings.append(self._reservoir._inputScaling)

                fitLosses.append(fitLoss)
                validationLosses.append(validationLoss)
            learningRates.append(learningRate)

        if verbose > 0:
            bar.finish()

        return (validationLosses, fitLosses, inputScalings, leakingRates, spectralRadiuses, learningRates)

    def optimizePenaltyForEvaluationError(self, trainingInputData, trainingOutputData, validationInputData, validationOutputData,
                                          optimizationLength, epochs=1, penalty=0.1, transientTime=0, verbose=1):

        self._validateReservoir()

        learningRate = self._learningRates[2]

        # initializations of arrays:
        if (len(trainingOutputData.shape) == 1):
            Ytarget = trainingOutputData[None, transientTime:].T
        else:
            Ytarget = trainingOutputData[transientTime:].T

        if (len(validationOutputData.shape) == 1):
            validationOutputData = validationOutputData[:, None]

        Ytarget = self._reservoir.out_inverse_activation(Ytarget)
        validationOutputData = self._reservoir.out_inverse_activation(validationOutputData)

        originalLearningRate = learningRate

        fitLosses = list()
        validationLosses = list()
        penalties = list()
        learningRates = list()

        penaltyDerivatives = B.zeros(optimizationLength)
        oldPenalty = penalty

        self._reservoir.fit(trainingInputData, trainingOutputData, transientTime=transientTime)
        oldLoss = hlp.loss(self._reservoir.predict(validationInputData), validationOutputData)

        evaluationEchoFunction = B.zeros((1 + self._reservoir.n_reservoir + self._reservoir.n_input, optimizationLength))
        x = self._reservoir.x

        for t, u in enumerate(validationInputData):
            u = u.reshape(-1, 1)
            self._reservoir.update(u)
            x = self._reservoir._x
            evaluationEchoFunction[:, t] = B.vstack((1, u, x)).squeeze()

        if (verbose > 0):
            bar = progressbar.ProgressBar(max_value=epochs, redirect_stdout=True, poll_interval=0.0001)
            bar.update(0)

        for epoch in range(epochs):
            if verbose > 0:
                bar.update(epoch)

            penaltyDerivative = self.derivationForPenalty(Ytarget, self._reservoir._X, penalty)

            for t in range(len(validationInputData)):
                predictionPoint = B.dot(self._reservoir._WOut, evaluationEchoFunction[:, t].reshape(-1, 1))
                error = (trainingOutputData[t] - predictionPoint).T
                penaltyDerivatives[t] = - B.dot(B.dot(error, penaltyDerivative), predictionPoint)

            penaltyGradient = sum(penaltyDerivatives)

            penaltyGradient = B.sign( penaltyGradient )

            penalty = penalty - learningRate * penaltyGradient

            self._reservoir.setPenalty(penalty)
            self._reservoir._calculateOutputMatrix()
            fitLoss = hlp.loss( self._reservoir.predict( trainingInputData), trainingOutputData )
            validationLoss = hlp.loss( self._reservoir.predict(validationInputData), validationOutputData )

            # this is the "when the error goes up, go back and divide learning rate by 2" mechanism
            if validationLoss > oldLoss:
                penalty = oldPenalty
                learningRate = learningRate / 2
            else:
                learningRate = min(learningRate*1.5, originalLearningRate)

                oldPenalty = penalty
                oldLoss = validationLoss
                fitLosses.append(fitLoss)
                validationLosses.append(validationLoss)
                penalties.append(penalty)
            learningRates.append(learningRate)

        if verbose > 0:
            bar.finish()

        return (validationLosses, fitLosses, penalties, learningRates)



    #####################################################################################################################################################

    def _derivationLrSrIsFsGenerative(self, oldDerivatives, ooldDerivatives, derivativesWout, W_uniform, W_in_uniform, W_fb_uniform, u, oldX, ooldX, oldy):
        doldSr, doldLr, doldIs, doldFs = oldDerivatives
        dooldXSr, dooldXLr, dooldXIs, dooldXFs = ooldDerivatives
        dWoutIs, dWoutLr, dWoutSr, dWoutFs = derivativesWout

        a = self._reservoir._leakingRate
        X = self._reservoir.calculateLinearNetworkTransmissions(u)
        u = B.vstack((1, u))
        dooldXSr, dooldXLr, dooldXIs, dooldXFs = B.vstack(( B.zeros_like(u), dooldXSr)), B.vstack(( B.zeros_like(u), dooldXLr)), B.vstack(( B.zeros_like(u), dooldXIs)), B.vstack(( B.zeros_like(u), dooldXFs))
        ooldX = B.vstack((u, ooldX))

        activationX = self._reservoir._activation(X)
        activationDerivationX = self._reservoir._activationDerivation(X)

        dyLr = B.dot(self._reservoir._WOut, dooldXLr) + B.dot(dWoutLr, ooldX)
        dyIs = B.dot(self._reservoir._WOut, dooldXIs) + B.dot(dWoutIs, ooldX)
        dySr = B.dot(self._reservoir._WOut, dooldXSr) + B.dot(dWoutSr, ooldX)
        dyFs = B.dot(self._reservoir._WOut, dooldXFs) + B.dot(dWoutFs, ooldX)

        dLr = (1-a) * doldLr - oldX + activationX + a * activationDerivationX * (B.dot(self._reservoir._W, doldLr) + B.dot(self._reservoir._WFeedback, dyLr))
        dSr = (1-a) * doldSr + a * activationDerivationX * (B.dot(self._reservoir._W, doldSr) + B.dot(W_uniform, oldX) + B.dot(self._reservoir._WFeedback, dySr))

        dIs = (1-a) * doldIs + a * activationDerivationX * (B.dot(self._reservoir._W, doldIs) + B.dot(W_in_uniform, u) + B.dot(self._reservoir._WFeedback, dyIs))

        dFs = (1-a) * doldFs + a * activationDerivationX * (B.dot(self._reservoir.W, doldFs) + B.dot(W_fb_uniform, oldy) + B.dot(self._reservoir.W_feedback, dyFs))

        return dLr, dSr, dIs, dFs


    def optimizeParameterForGenerativeValidationError(self, trainingInputData, trainingOutputData, validationOutputData,
                                            epochs=1, transientTime=None, verbose=1):

        self._validateReservoir()

        learningRate = self._learningRates[1]

        if not np.isscalar(self._reservoir._inputScaling):
            raise ValueError(
                "Only penalty optimization is supported for a multiple input scalings at the moment. We are working on it.")

        # calculate stuff
        trainLength = trainingInputData.shape[0]
        optimizationLength = validationOutputData.shape[0]
        if transientTime is None:
            transientTime = self._reservoir.estimateTransientTime(trainingInputData, trainingOutputData)

        # initializations of arrays:
        if (len(trainingOutputData.shape) == 1):
            Ytarget = trainingOutputData[None, transientTime:].T
        else:
            Ytarget = trainingOutputData[transientTime:].T

        if (len(validationOutputData.shape) == 1):
            validationOutputData = validationOutputData[:, None]

        Ytarget = self._reservoir.out_inverse_activation(Ytarget)
        validationOutputData = self._reservoir.out_inverse_activation(validationOutputData)

        originalLearningRate = learningRate

        # initializations for plotting parameter and losses at the end
        inputScalings = list()
        feedbackScalings = list()
        leakingRates = list()
        spectralRadiuses = list()
        fitLosses = list()
        validationLosses = list()
        learningRates = list()

        # initializations for arrays which collect all the gradients of the error of the single time steps, which get add at the end
        srGradients = B.zeros(optimizationLength)
        lrGradients = B.zeros(optimizationLength)
        isGradients = B.zeros(optimizationLength)
        fsGradients = B.zeros(optimizationLength)

        # collecting the single derivatives  - > this is the derivation of design matrix when filled
        srGradientsMatrix = B.zeros(
            (self._reservoir.n_reservoir + self._reservoir.n_input + 1, trainLength - transientTime))
        lrGradientsMatrix = B.zeros(
            (self._reservoir.n_reservoir + self._reservoir.n_input + 1, trainLength - transientTime))
        isGradientsMatrix = B.zeros(
            (self._reservoir.n_reservoir + self._reservoir.n_input + 1, trainLength - transientTime))
        fsGradientsMatrix = B.zeros(
            (self._reservoir.n_reservoir + self._reservoir.n_input + 1, trainLength - transientTime))

        # initialize self.designMatrix and self.W_out
        self._reservoir.fit(trainingInputData, trainingOutputData, transientTime=transientTime)
        oldLoss = hlp.loss(self._reservoir.generate(optimizationLength, trainingInputData[-1]), validationOutputData)

        # initialize variables for the "when the error goes up, go back and divide learning rate by 2" mechanism
        oldSR = self._reservoir._spectralRadius
        oldLR = self._reservoir._leakingRate
        oldIS = self._reservoir._inputScaling
        oldFS = self._reservoir._feedbackScaling


        # Calculate uniform matrices
        W_uniform = self._reservoir._W / self._reservoir._spectralRadius
        W_in_uniform = self._reservoir._WInput / self._reservoir._inputScaling
        W_fb_uniform = self._reservoir._WFeedback / self._reservoir._feedbackScaling

        if (verbose > 0):
            bar = progressbar.ProgressBar(max_value=epochs, redirect_stdout=True, poll_interval=0.0001)
            bar.update(0)

        for epoch in range(epochs):
            if verbose > 0:
                bar.update(epoch)

            # initialize del x / del (rho, alpha, s_in) and reservoir state itself
            derivationSpectralRadius = B.empty((self._reservoir.n_reservoir, 1))
            derivationLeakingRate = B.empty((self._reservoir.n_reservoir, 1))
            derivationInputScaling = B.empty((self._reservoir.n_reservoir, 1))
            derivationFeedbackScaling = B.empty((self._reservoir.n_reservoir, 1))

            oldDerivationSpectralRadius = B.zeros((self._reservoir.n_reservoir, 1))
            oldDerivationLeakingRate = B.zeros((self._reservoir.n_reservoir, 1))
            oldDerivationInputScaling = B.zeros((self._reservoir.n_reservoir, 1))
            oldDerivationFeedbackScaling = B.zeros((self._reservoir.n_reservoir, 1))

            ooldDerivationSpectralRadius = B.zeros((self._reservoir.n_reservoir, 1))
            ooldDerivationLeakingRate = B.zeros((self._reservoir.n_reservoir, 1))
            ooldDerivationInputScaling = B.zeros((self._reservoir.n_reservoir, 1))
            ooldDerivationFeedbackScaling = B.zeros((self._reservoir.n_reservoir, 1))

            x = B.zeros((self._reservoir.n_reservoir, 1))

            # go thorugh the train length (e.g. the time, on which W_out gets calculated)
            for t, u in enumerate(trainingInputData):
                u = u.reshape(-1, 1)
                oldx = x
                x = self._reservoir._X[2:, t]

                # calculate the del /x del (rho, alpha, s_in)
                derivationLeakingRate, derivationSpectralRadius, derivationInputScaling, derivationFeedbackScaling = self._derivationLrSrIsFs(
                    (oldDerivationSpectralRadius, oldDerivationLeakingRate, oldDerivationInputScaling, oldDerivationFeedbackScaling),
                    (ooldDerivationSpectralRadius, ooldDerivationLeakingRate, ooldDerivationInputScaling, ooldDerivationFeedbackScaling),
                    W_uniform, W_in_uniform, u, oldx, generative=True, W_fb_uniform = W_fb_uniform, oldy = trainingInputData[:,t+1])

                ooldDerivationSpectralRadius, ooldDerivationLeakingRate, ooldDerivationInputScaling, ooldDerivationFeedbackScaling = oldDerivationSpectralRadius, oldDerivationLeakingRate, oldDerivationInputScaling, oldDerivationFeedbackScaling
                oldDerivationSpectralRadius, oldDerivationLeakingRate, oldDerivationInputScaling, oldDerivationFeedbackScaling = derivationLeakingRate, derivationSpectralRadius, derivationInputScaling, derivationFeedbackScaling

                if t >= transientTime:
                    # concatenate with zeros (for the derivatives of the input and the 1, which are always 0)
                    srGradientsMatrix[:, t - transientTime] = B.concatenate(
                        (B.zeros(self._reservoir.n_input + 1), derivationSpectralRadius[:, 0]), axis=0)
                    lrGradientsMatrix[:, t - transientTime] = B.concatenate(
                        (B.zeros(self._reservoir.n_input + 1), derivationLeakingRate[:, 0]), axis=0)
                    isGradientsMatrix[:, t - transientTime] = B.concatenate(
                        (B.zeros(self._reservoir.n_input + 1), derivationInputScaling[:, 0]), axis=0)
                    fsGradientsMatrix[:, t - transientTime] = B.concatenate(
                        (B.zeros(self._reservoir.n_input + 1), derivationFeedbackScaling[:, 0]), axis=0)

            # add to matrix
            WoutPrimeSR = self.derivationWoutForP(Ytarget, self._reservoir._X, srGradientsMatrix)
            WoutPrimeLR = self.derivationWoutForP(Ytarget, self._reservoir._X, lrGradientsMatrix)
            WoutPrimeIS = self.derivationWoutForP(Ytarget, self._reservoir._X, isGradientsMatrix)
            WoutPrimeFS = self.derivationWoutForP(Ytarget, self._reservoir._X, fsGradientsMatrix)

            y =  trainingOutputData[-1]

            # this time go through validation length
            for t, ytarget in enumerate(validationOutputData):
                u = y
                ooldx = oldx
                oldx = x
                self._reservoir.update(u)
                x = self._reservoir._x

                # calculate error at given time step
                y = B.dot(self._reservoir._WOut, B.vstack((1, u, x)))
                error = (ytarget - y).T

                # calculate the del /x del (rho, alpha, s_in)
                derivationLeakingRate, derivationSpectralRadius, derivationInputScaling, derivationFeedbackScaling = self._derivationLrSrIsFsGenerative(
                    (oldDerivationSpectralRadius, oldDerivationLeakingRate, oldDerivationInputScaling, oldDerivationFeedbackScaling),
                    (ooldDerivationSpectralRadius, ooldDerivationLeakingRate, ooldDerivationInputScaling, ooldDerivationFeedbackScaling),
                    (WoutPrimeSR, WoutPrimeLR, WoutPrimeIS, WoutPrimeFS),
                    W_uniform, W_in_uniform, W_fb_uniform, u, oldx, ooldx, oldy)

                ooldDerivationSpectralRadius, ooldDerivationLeakingRate, ooldDerivationInputScaling, ooldDerivationFeedbackScaling = oldDerivationSpectralRadius, oldDerivationLeakingRate, oldDerivationInputScaling, oldDerivationFeedbackScaling
                oldDerivationSpectralRadius, oldDerivationLeakingRate, oldDerivationInputScaling, oldDerivationFeedbackScaling = derivationLeakingRate, derivationSpectralRadius, derivationInputScaling, derivationFeedbackScaling
                oldy = y

                # concatenate derivations with 0
                derivationConcatinationSpectralRadius = B.concatenate(
                    (B.zeros(self._reservoir.n_input + 1), derivationSpectralRadius[:, 0]), axis=0)
                derivationConcatinationLeakingRate = B.concatenate(
                    (B.zeros(self._reservoir.n_input + 1), derivationLeakingRate[:, 0]), axis=0)
                derivationConcatinationInputScaling = B.concatenate(
                    (B.zeros(self._reservoir.n_input + 1), derivationInputScaling[:, 0]), axis=0)
                derivationConcatinationFeedbackScaling = B.concatenate(
                    (B.zeros(self._reservoir.n_input + 1), derivationFeedbackScaling[:, 0]), axis=0)

                # calculate gradients
                gradientSR = B.dot(-error,
                                   B.dot(self._reservoir._WOut, derivationConcatinationSpectralRadius) + B.dot(
                                       WoutPrimeSR, B.vstack((1, u, x))[:, 0]))
                srGradients[t] = gradientSR
                gradientLR = B.dot(-error,
                                   B.dot(self._reservoir._WOut, derivationConcatinationLeakingRate) + B.dot(
                                       WoutPrimeLR, B.vstack((1, u, x))[:, 0]))
                lrGradients[t] = gradientLR
                gradientIS = B.dot(-error,
                                   B.dot(self._reservoir._WOut, derivationConcatinationInputScaling) + B.dot(
                                       WoutPrimeIS, B.vstack((1, u, x))[:, 0]))
                isGradients[t] = gradientIS
                gradientFS = B.dot(-error,
                                   B.dot(self._reservoir._WOut, derivationConcatinationFeedbackScaling) + B.dot(
                                       WoutPrimeFS, B.vstack((1, u, x))[:, 0]))
                fsGradients[t] = gradientFS

            # sum up the gradients del error / del (rho, alpha, s_in) to get final gradient
            gradientSR = sum(srGradients)
            gradientLR = sum(lrGradients)
            gradientIS = sum(isGradients)
            gradientFS = sum(fsGradients)

            # normalize length of gradient to 1
            gradientVectorLength = B.sqrt(gradientSR ** 2 + gradientLR ** 2 + gradientIS ** 2 + gradientFS ** 2)

            gradientSR /= gradientVectorLength
            gradientLR /= gradientVectorLength
            gradientIS /= gradientVectorLength
            gradientFS /= gradientVectorLength

            # update spectral radius
            self._reservoir.setSpectralRadius(self._reservoir._spectralRadius - learningRate * gradientSR)

            # update leaking rate
            self._reservoir.setLeakingRate(self._reservoir._leakingRate - learningRate * gradientLR)

            # update input scaling
            self._reservoir.setInputScaling(self._reservoir._inputScaling - learningRate * gradientIS)

            # update input scaling
            self._reservoir.setFeedbackScaling(self._reservoir._feedbackScaling - learningRate * gradientFS)

            # calculate the errors and update the self.designMatrix and the W_out
            fitLoss = self._reservoir.fit(trainingInputData, trainingOutputData, transientTime=transientTime)
            validationLoss = hlp.loss(self._reservoir.generate(optimizationLength, trainingInputData[-1]), validationOutputData)

            # this is the "when the error goes up, go back and divide learning rate by 2" mechanism
            if validationLoss > oldLoss:
                self._reservoir.setSpectralRadius(oldSR)
                self._reservoir.setLeakingRate(oldLR)
                self._reservoir.setInputScaling(oldIS)
                self._reservoir.setFeedbackScaling(oldFS)
                learningRate = learningRate / 2
            else:
                learningRate = min(learningRate*1.5, originalLearningRate)

                oldSR = self._reservoir._spectralRadius
                oldLR = self._reservoir._leakingRate
                oldIS = self._reservoir._inputScaling
                oldFS = self._reservoir._feedbackScaling
                oldLoss = validationLoss

                spectralRadiuses.append(self._reservoir._spectralRadius)
                leakingRates.append(self._reservoir._leakingRate)
                inputScalings.append(self._reservoir._inputScaling)
                feedbackScalings.append(self._reservoir._feedbackScaling)

                fitLosses.append(fitLoss)
                validationLosses.append(validationLoss)

            learningRates.append(learningRate)

        if verbose > 0:
            bar.finish()

        return (validationLosses, fitLosses, inputScalings, leakingRates, spectralRadiuses, learningRates)


"""
    Utility methods which are used for the ESNs.
"""

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as animation

"""
    Calculates the mutual information between the two signals x and y.
"""
def calculate_mutualinformation(x, y, bins):
    pxy, _, _ = np.histogram2d(x, y, bins)
    px, _, = np.histogram(x, bins)
    py, _, = np.histogram(y, bins)

    pxy = pxy/np.sum(pxy)
    px = px/np.sum(px)
    py = py/np.sum(py)

    pxy = pxy[np.nonzero(pxy)]
    px = px[np.nonzero(px)]
    py = py[np.nonzero(py)]

    hxy = -np.sum(pxy*np.log2(pxy))
    hx = -np.sum(px*np.log2(px))
    hy = -np.sum(py*np.log2(py))

    MI = hx+hy-hxy

    return MI

"""
    Tries to calculate the input scaling of an ESN by using the mutual information.
"""
def calculate_esn_mi_input_scaling(input_data, output_data):
    if len(input_data) != len(output_data):
        raise ValueError("input_data and output_data do not have the same length -  {0} vs. {1}".format(len(input_data), len(output_data)))

    #Scott's rule to calculate nbins
    std_output = np.std(output_data)
    nbins = int(np.ceil(2.0/(3.5*std_output/np.power(len(input_data), 1.0/3.0))))

    mi = np.zeros(input_data.shape[1])
    for i in range(len(mi)):
        mi[i] = calculate_mutualinformation(input_data[:, i], output_data, nbins)
    scaling = mi / np.max(mi)

    return scaling


"""
    Estimates autocorrelation of given np array x
"""
def autocorrelation(self, x):
    n = x.shape[0]
    variance = B.var(x)
    x = x - B.mean(x)
    r = B.correlate(x, x, mode='full')[-n:]
    assert B.allclose(r, np.array([(x[:n - k] * x[-(n - k):]).sum() for k in range(n)]))
    result = r / (variance * (np.arange(n, 0, -1)))
    return result

"""
    Calculates SWD (sliding window difference) with the specified intervall using the first `interval` entries of the series as the window
    Returns a tuple consisting of the point of minimum and the whole SWD series
"""
def SWD(series, intervall):
    differences = np.zeros(series.shape[0] - 2 * intervall)
    reference_series = series[:intervall]
    for i in range(intervall, series.shape[0] - intervall):
        differences[i - intervall] = np.sum(np.abs(reference_series - series[i:i + intervall]))

    return np.argmin(differences) + intervall, differences


def loss(prediction, target):
    return np.mean( ( prediction - target ) ** 2 )



"""
    Basic implementation of an ESN.
"""

#from __future__ import absolute_import

import numpy as np
import numpy.random as rnd
import dill as pickle
import scipy as sp
import progressbar
from . import helper as hp

#import backend as B

from . import backend as B

class BaseESN(object):
    def __init__(self, n_input, n_reservoir, n_output,
                 spectralRadius=1.0, noiseLevel=0.01, inputScaling=None,
                 leakingRate=1.0, feedbackScaling = 1.0, reservoirDensity=0.2, randomSeed=None,
                 out_activation=lambda x: x, out_inverse_activation=lambda x: x,
                 weightGeneration='naive', bias=1.0, outputBias=1.0, outputInputScaling=1.0,
                 feedback=False, input_density=1.0, activation = B.tanh, activationDerivation=lambda x: 1.0/B.cosh(x)**2):

        self.n_input = n_input
        self.n_reservoir = n_reservoir
        self.n_output = n_output

        self._spectralRadius = spectralRadius
        self._noiseLevel = noiseLevel
        self._reservoirDensity = reservoirDensity
        self._leakingRate = leakingRate
        self._feedbackScaling = feedbackScaling
        self.input_density = input_density
        self._activation = activation
        self._activationDerivation = activationDerivation

        if inputScaling is None:
            self._inputScaling = 1.0
        if np.isscalar(self._inputScaling):
            inputScaling = B.ones(n_input) * self._inputScaling
        else:
            if len(self._inputScaling) != self.n_input:
                raise ValueError("Dimension of inputScaling ({0}) does not match the input data dimension ({1})".format(len(self._inputScaling), n_input))
            self._inputScaling = inputScaling

        self._expandedInputScaling = B.vstack((B.array(1.0), inputScaling.reshape(-1,1))).flatten()

        self.out_activation = out_activation
        self.out_inverse_activation = out_inverse_activation

        if randomSeed is not None:
            rnd.seed(randomSeed)

        self._bias = bias
        self._outputBias = outputBias
        self._outputInputScaling = outputInputScaling
        self._createReservoir(weightGeneration, feedback)


    def setSpectralRadius(self, newSpectralRadius):
        self._W = self._W * ( newSpectralRadius / self._spectralRadius )
        self._spectralRadius = newSpectralRadius
        #TODO numerical instability

    def setLeakingRate(self, newLeakingRate):
        self._leakingRate = newLeakingRate

    def setInputScaling(self, newInputScaling):
        inputScaling = B.ones(self.n_input) * self._inputScaling
        self._expandedInputScaling = B.vstack((B.array(1.0), inputScaling.reshape(-1, 1))).flatten()
        self._WInput = self._WInput * ( self._expandedInputScaling / self._inputScaling )
        self._inputScaling = newInputScaling

    def setFeedbackScaling(self, newFeedbackScaling):
        self._WFeedback = self._WFeedback * ( newFeedbackScaling / self._feedbackScaling)
        self._feedbackScaling = newFeedbackScaling


    def resetState(self):
        self._x = B.zeros_like(self._x)

    def propagate(self, inputData, outputData=None, transientTime=0, verbose=0, x=None, steps="auto"):
        if x is None:
            x = self._x

        inputLength = steps
        if inputData is None:
            if outputData is not None: 
                inputLength = len(outputData)
        else:
            inputLength = len(inputData)
        if inputLength == "auto":
            raise ValueError("inputData and outputData are both None. Therefore, steps must not be `auto`.")

        # define states' matrix
        X = B.zeros((1 + self.n_input + self.n_reservoir, inputLength - transientTime))

        if (verbose > 0):
            bar = progressbar.ProgressBar(max_value=inputLength, redirect_stdout=True, poll_interval=0.0001)
            bar.update(0)

        if self._WFeedback is None:
            #do not distinguish between whether inputData is None or not, as the feedback has been disabled
            #therefore, the input has to be anything but None

            for t in range(inputLength):
                u = self.update(inputData[t], x=x)
                if (t >= transientTime):
                    #add valueset to the states' matrix
                    X[:,t-transientTime] = B.vstack((B.array(self._outputBias), self._outputInputScaling*u, x))[:,0]
                if (verbose > 0):
                    bar.update(t)
        else:
            if outputData is None:
                Y = B.empty((inputLength-transientTime, self.n_output))

            previousOutputData = B.zeros((1, self.n_output))

            if inputData is None:
                for t in range(inputLength):
                    self.update(None, previousOutputData, x=x)
                    if (t >= transientTime):
                        #add valueset to the states' matrix
                        X[:,t-transientTime] = B.vstack((B.array(self._outputBias), x))[:,0]
                    if outputData is None:
                        #calculate the prediction using the trained model
                        if (self._solver in ["sklearn_auto", "sklearn_lsqr", "sklearn_sag", "sklearn_svd"]):
                            previousOutputData = self._ridgeSolver.predict(B.vstack((B.array(self._outputBias), self._x)).T)
                        else:
                            previousOutputData = B.dot(self._WOut, B.vstack((B.array(self._outputBias), self._x)))
                        if t >= transientTime:
                            Y[t-transientTime, :] = previousOutputData
                    else:
                        previousOutputData = outputData[t]
                
                    if (verbose > 0):
                        bar.update(t)
            else:
                for t in range(inputLength):
                    u = self.update(inputData[t], previousOutputData, x=x)
                    if (t >= transientTime):
                        #add valueset to the states' matrix
                        X[:,t-transientTime] = B.vstack((B.array(self._outputBias), self._outputInputScaling*u, x))[:,0]
                    if outputData is None:
                        #calculate the prediction using the trained model
                        if (self._solver in ["sklearn_auto", "sklearn_lsqr", "sklearn_sag", "sklearn_svd"]):
                            previousOutputData = self._ridgeSolver.predict(B.vstack((B.array(self._outputBias), self._outputInputScaling*u, self._x)).T)
                        else:
                            previousOutputData = B.dot(self._WOut, B.vstack((B.array(self._outputBias), self._outputInputScaling*u, self._x)))
                        Y[t, :] = previousOutputData
                    else:
                        previousOutputData = outputData[t]
                
                    if (verbose > 0):
                        bar.update(t)
                                 
        if (verbose > 0):
            bar.finish()

        if self._WFeedback is not None and outputData is None:
            return X, Y
        else:
            return X

    """
        Generates a random rotation matrix, used in the SORM initilization (see http://ftp.math.uni-rostock.de/pub/preprint/2012/pre12_01.pdf)
    """
    def create_random_rotation_matrix(self):
        h = rnd.randint(low=0, high=self.n_reservoir)
        k = rnd.randint(low=0, high=self.n_reservoir)

        phi = rnd.rand(1)*2*np.pi

        Q = B.identity(self.n_reservoir)
        Q[h, h] = np.cos(phi)
        Q[k, k] = np.cos(phi)

        Q[h, k] = -np.sin(phi)
        Q[k, h] = np.sin(phi)

        return Q

    """
        Internal method to create the matrices W_in, W and W_fb of the ESN
    """
    def _createReservoir(self, weightGeneration, feedback=False, verbose=False):
        #naive generation of the matrix W by using random weights
        if weightGeneration == 'naive':
            #random weight matrix from -0.5 to 0.5
            self._W = B.array(rnd.rand(self.n_reservoir, self.n_reservoir) - 0.5)

            #set sparseness% to zero
            mask = rnd.rand(self.n_reservoir, self.n_reservoir) > self._reservoirDensity
            self._W[mask] = 0.0

            _W_eigenvalues = B.abs(B.eigenval(self._W)[0])
            self._W *= self._spectralRadius / B.max(_W_eigenvalues)

        #generation using the SORM technique (see http://ftp.math.uni-rostock.de/pub/preprint/2012/pre12_01.pdf)
        elif weightGeneration == "SORM":
            self._W = B.identity(self.n_reservoir)

            number_nonzero_elements = self._reservoirDensity * self.n_reservoir * self.n_reservoir
            i = 0

            while np.count_nonzero(self._W) < number_nonzero_elements:
                i += 1
                Q = self.create_random_rotation_matrix()
                self._W = Q.dot(self._W)
            
            self._W *= self._spectralRadius

        #generation using the proposed method of Yildiz
        elif weightGeneration == 'advanced':
            #two create W we must follow some steps:
            #at first, create a W = |W|
            #make it sparse
            #then scale its spectral radius to rho(W) = 1 (according to Yildiz with x(n+1) = (1-a)*x(n)+a*f(...))
            #then change randomly the signs of the matrix

            #random weight matrix from 0 to 0.5

            self._W = B.array(rnd.rand(self.n_reservoir, self.n_reservoir) / 2)

            #set sparseness% to zero
            mask = B.rand(self.n_reservoir, self.n_reservoir) > self._reservoirDensity
            self._W[mask] = 0.0

            from scipy.sparse.linalg.eigen.arpack.arpack import ArpackNoConvergence
            #just calculate the largest EV - hopefully this is the right code to do so...
            try:
                #this is just a good approximation, so this code might fail
                _W_eigenvalue = B.max(np.abs(sp.sparse.linalg.eigs(self._W, k=1)[0]))
            except ArpackNoConvergence:
                #this is the safe fall back method to calculate the EV
                _W_eigenvalue = B.max(B.abs(sp.linalg.eigvals(self._W)))
            #_W_eigenvalue = B.max(B.abs(np.linalg.eig(self._W)[0]))

            self._W *= self._spectralRadius / _W_eigenvalue

            if verbose:
                M = self._leakingRate*self._W + (1 - self._leakingRate)*np.identity(n=self._W.shape[0])
                M_eigenvalue = B.max(B.abs(B.eigenval(M)[0]))#np.max(np.abs(sp.sparse.linalg.eigs(M, k=1)[0]))
                print("eff. spectral radius: {0}".format(M_eigenvalue))

            #change random signs
            random_signs = B.power(-1, rnd.random_integers(self.n_reservoir, self.n_reservoir))

            self._W = B.multiply(self._W, random_signs)
        elif weightGeneration == 'custom':
            pass
        else:
            raise ValueError("The weightGeneration property must be one of the following values: naive, advanced, SORM, custom")

        #check of the user is really using one of the internal methods, or wants to create W by his own
        if (weightGeneration != 'custom'):
            #random weight matrix for the input from -0.5 to 0.5
            self._WInput = B.rand(self.n_reservoir, 1 + self.n_input)-0.5

            #scale the input_density to prevent saturated reservoir nodes
            if (self.input_density != 1.0):
                #make the input matrix as dense as requested
                input_topology = (np.ones_like(self._WInput) == 1.0)
                nb_non_zero_input = int(self.input_density * self.n_input)
                for n in range(self.n_reservoir):
                    input_topology[n][rnd.permutation(np.arange(1+self.n_input))[:nb_non_zero_input]] = False

                self._WInput[input_topology] = 0.0

            self._WInput = self._WInput * self._expandedInputScaling

        #create the optional feedback matrix
        if feedback:
            self._WFeedback = B.rand(self.n_reservoir, 1 + self.n_output) - 0.5
            self._WFeedback *= self._feedbackScaling
        else:
            self._WFeedback = None


    def calculateLinearNetworkTransmissions(self, u, x=None):
        if x is None:
            x = self._x

        return B.dot(self._WInput, B.vstack((B.array(self._bias), u))) + B.dot(self._W, x)

    """
        Updates the inner states. Returns the UNSCALED but reshaped input of this step.
    """
    def update(self, inputData, outputData=None, x=None):
        if x is None:
            x = self._x

        if self._WFeedback is None:
            #reshape the data
            u = inputData.reshape(self.n_input, 1)

            #update the states
            transmission = self.calculateLinearNetworkTransmissions(u, x)
            x *= (1.0-self._leakingRate)
            x += self._leakingRate * self._activation(transmission + (B.rand()-0.5)*self._noiseLevel)
        
            return u

        else:
            #the input is allowed to be "empty" (size=0)
            if self.n_input != 0:
                #reshape the data
                u = inputData.reshape(self.n_input, 1)
                outputData = outputData.reshape(self.n_output, 1)

                #update the states
                transmission = self.calculateLinearNetworkTransmissions(u, x)
                x *= (1.0-self._leakingRate)
                x += self._leakingRate*self._activation(transmission +
                     B.dot(self._WFeedback, B.vstack((B.array(self._outputBias), outputData))) + (B.rand()-0.5)*self._noiseLevel)

                return u
            else:
                #reshape the data
                outputData = outputData.reshape(self.n_output, 1)
                #update the states
                transmission = B.dot(self._W, x)
                x *= (1.0-self._leakingRate)
                x += self._leakingRate*self._activation(transmission + B.dot(self._WFeedback, B.vstack((B.array(self._outputBias), outputData))) +
                     (B.rand()-0.5)*self._noiseLevel)

                return np.empty((0, 1))

    def calculateTransientTime(self, inputs, outputs, epsilon, proximityLength = None):
        # inputs: input of reserovoir
        # outputs: output of reservoir
        # epsilon: given constant
        # proximity length: number of steps for which all states have to be epsilon close to declare convergance
        # initializes two initial states as far as possible from each other in [-1,1] regime and tests when they converge-> this is transient time

        length = inputs.shape[0] if inputs is not None else outputs.shape[0]
        if proximityLength is None:
            proximityLength = int(length * 0.1)
            if proximityLength < 3:
                proximityLength = 3

        initial_x = B.empty((2, self.n_reservoir, 1))
        initial_x[0] = - B.ones((self.n_reservoir, 1))
        initial_x[1] = B.ones((self.n_reservoir, 1))

        countedConsecutiveSteps = 0
        length = inputs.shape[0] if inputs is not None else outputs.shape[0]
        for t in range(length):
            if B.max(B.ptp(initial_x, axis=0)) < epsilon:
                if countedConsecutiveSteps >= proximityLength:
                    return t - proximityLength
                else:
                    countedConsecutiveSteps += 1
            else:
                countedConsecutiveSteps = 0

            u = inputs[t].reshape(-1, 1) if inputs is not None else None
            o = outputs[t].reshape(-1, 1) if outputs is not None else None
            for i in range(initial_x.shape[0]):
                self.update(u, o, initial_x[i])

        #transient time could not be determined
        raise ValueError("Transient time could not be determined - maybe the proximityLength is too big.")       

 
    def reduceTransientTime(self, inputs, outputs, initialTransientTime, epsilon = 1e-3, proximityLength = 50):
        # inputs: input of reserovoir
        # outputs: output of reservoir
        # epsilon: given constant
        # proximity length: number of steps for which all states have to be epsilon close to declare convergance
        # initialTransientTime: transient time with calculateTransientTime() method estimated
        # finds initial state with lower transient time and sets internal state to this state
        # returns the new transient time by calculating the convergence time of initial states found with SWD and Equilibrium method
 
        def getEquilibriumState(inputs, outputs, epsilon = 1e-3):
            # inputs: input of reserovoir
            # outputs: output of reservoir
            # epsilon: given constant
            # returns the equilibrium state when esn is fed with the first state of input
            x = B.empty((2, self.n_reservoir, 1))
            while not B.max(B.ptp(x, axis=0)) < epsilon:
                x[0] = x[1]
                u = inputs[0].reshape(-1, 1) if inputs is not None else None
                o = outputs[0].reshape(-1, 1) if outputs is not None else None
                self.update(u, o, x[1])

            return x[1]

        def getStateAtGivenPoint(inputs, outputs, targetTime):
            # inputs: input of reserovoir
            # outputs: output of reservoir
            # targetTime: time at which the state is wanted
            # propagates the inputs/outputs till given point in time and returns the state of the reservoir at this point
            x = B.zeros((self.n_reservoir, 1))

            length = inputs.shape[0] if inputs is not None else outputs.shape[0]
            length = min(length, targetTime)
            for t in range(length):
                u = inputs[t].reshape(-1, 1) if inputs is not None else None
                o = outputs[t].reshape(-1, 1) if outputs is not None else None
                self.update(u, o, x)
        
            return x

        length = inputs.shape[0] if inputs is not None else outputs.shape[0]
        if proximityLength is None:
            proximityLength = int(length * 0.1)
            if proximityLength < 3:
                proximityLength = 3
     
        x = B.empty((2, self.n_reservoir, 1))
        equilibriumState = getEquilibriumState(inputs, outputs)
     
        if inputs is None:
            swdPoint, _ = hp.SWD(outputs, int(initialTransientTime*0.8))
        else:
            swdPoint, _ = hp.SWD(inputs, int(initialTransientTime * 0.8))
        
        swdState = getStateAtGivenPoint(inputs, outputs, swdPoint)
        
        x[0] = equilibriumState
        x[1] = swdState
        
        transientTime = 0

        countedConsecutiveSteps = 0
        for t in range(length):
            if B.max(B.ptp(x, axis=0)) < epsilon:
                countedConsecutiveSteps += 1
                if countedConsecutiveSteps > proximityLength:
                    transientTime = t - proximityLength
                    break
            else:
                countedConsecutiveSteps = 0

            u = inputs[t].reshape(-1, 1) if inputs is not None else None
            o = outputs[t].reshape(-1, 1) if outputs is not None else None
            for i in range(x.shape[0]):
                self.update(u, o, x[i])

        self._x = x[0]
        return transientTime

    """
        Saves the ESN by pickling it.
    """
    def save(self, path):
        f = open(path, "wb")
        pickle.dump(self, f)
        f.close()

    """
        Loads a previously pickled ESN.
    """
    def load(path):
        f = open(path, "rb")
        result = pickle.load(f)
        f.close()
        return result




class Pipeline(object):
    def __init__(self, **items):
        self.items = items

    def fit(self, trainingInput, trainingOutput, validationInput, validationOutput, verbose=1):
        for item in self.items:
            item.fit(trainingInput, trainingOutput, validationInput, validationOutput, verbose)


import numpy as np

class OneHotEncoder(object):
    def __init__(self, classes=None, handleUnknown="error"):
        self.classes = classes
        self.handleUnknown = handleUnknown

    def fit(self, data):
        if len(data.shape) == 1:
            data = data.reshape(-1, 1)
        
        if data.shape[1] != 1:
            raise ValueError("The data has to one dimensional, so it either has to have the shape (*) or (*, 1).")

        if self.classes is None:
            #get unique classes
            self.classes = np.unique(data)


    def transform(self, data):
        result = np.zeros((len(data), len(self.classes)))
        for i in range(len(data)):
            result[i, np.searchsorted(self.classes, data[i])] = 1.0

        return result

    def fit_transform(self, data):
        self.fit(data)
        return self.transform(data)


import cupy as cp

def add(x, y):
	return cp.add(x, y)

def substract(x, y):
	return cp.substract(x, y)

def dot(x, y):
	return cp.dot(x, y)

def multiply(x, y):
	return cp.multiply(x, y)

def eigenval(x):
	import numpy as np
	np_x = cp.asnumpy(x)
	results = np.linalg.eig(np_x)
	return cp.array(results[0]), cp.array([results[1]])

def array(x):
	return cp.array(x)

def inv(x):
	return cp.linalg.inv(x)

def pinv(x):
	return cp.linalg.pinv(x)

def arctan(x):
	return cp.arctan(x)

def vstack(x):
	return cp.vstack(x)

def abs(x):
	return cp.abs(x)

def max(x):
	return cp.max(x)

def ones(x):
	return cp.ones(x)

def zeros(x):
	return cp.zeros(x)

def empty(x):
	return cp.empty(x)

def mean(x, axis=None):
	return cp.mean(x, axis)

def sqrt(x):
	return cp.sqrt(x)

def identity(x):
	return cp.identity(x)

def rand(*x):
	return cp.random.rand(*x)

def power(x, y):
	return cp.power(x, y)

def exp(x):
	return cp.exp(x)

def cosh(x):
    return cp.cosh(x)  

def log(x):
    return cp.log(x)

def tanh(x):
	return cp.tanh(x)

def concatenate(tuple, axis=0):
	return cp.concatenate(tuple, axis=axis)

def sign(x):
	return cp.sign(x)

def argmax(x, axis):
    return cp.argmax(x, axis)

def zeros_like(x):
    return cp.zeros_like(x)


import sys
import os
import json

#default backend
_BACKEND = 'numpy'

#inspired by Keras backend handling

# Obtain easyesn base directory path: either ~/.easyesn or /tmp.
_easyesn_base_dir = os.path.expanduser('~')
if not os.access(_easyesn_base_dir, os.W_OK):
    _easyesn_base_dir = '/tmp'
_easyesn_dir = os.path.join(_easyesn_base_dir, '.easyesn')

# Attempt to read easyesn config file.
_config_path = os.path.expanduser(os.path.join(_easyesn_dir, 'easyesn.json'))
if os.path.exists(_config_path):
    try:
        _config = json.load(open(_config_path))
    except ValueError:
        _config = {}
    _backend = _config.get('backend', _BACKEND)
    assert _backend in {'numpy', 'np', 'cupy', 'cp'}
    
    if _backend == "cp":
        _backend = "cupy"
    if _backend == "np":
        _backend = "numpy"

    _BACKEND = _backend

# Save config file, if possible.
if not os.path.exists(_easyesn_dir):
    try:
        os.makedirs(_easyesn_dir)
    except OSError:
        pass

if not os.path.exists(_config_path):
    _config = {
        'backend': _BACKEND
    }
    try:
        with open(_config_path, 'w') as f:
            f.write(json.dumps(_config, indent=4))
    except IOError:
        # Except permission denied.
        pass

# Set backend based on EASYESN_BACKEND flag, if applicable.
if 'EASYESN_BACKEND' in os.environ:
    _backend = os.environ['EASYESN_BACKEND']
    assert _backend in {'numpy', 'np', 'cupy', 'cp'}
    
    if _backend == "cp":
        _backend = "cupy"
    if _backend == "np":
        _backend = "numpy"

    _BACKEND = _backend

if _BACKEND == 'cupy':
    sys.stderr.write('Using CuPy backend\n')
    from .cupyBackend import *
elif _BACKEND == 'numpy':
    sys.stderr.write('Using Numpy backend.\n')
    from .numpyBackend import *
else:
    raise ValueError('Unknown backend: ' + str(_BACKEND))

def backendName():
    """Publicly accessible method
    for determining the current backend.
    # Returns
        String, the name of the backend which is currently used.
    """
    return _BACKEND


from .PredictionESN import PredictionESN
from .RegressionESN import RegressionESN
from .ClassificationESN import ClassificationESN
from .SpatioTemporalESN import SpatioTemporalESN

from .OneHotEncoder import OneHotEncoder



import numpy as np

def add(x, y):
	return np.add(x, y)

def substract(x, y):
	return np.substract(x, y)

def dot(x, y):
	return np.dot(x, y)

def multiply(x, y):
	return np.multiply(x, y)

def eigenval(x):
	return np.linalg.eig(x)

def array(x):
	return np.array(x)

def inv(x):
	return np.linalg.inv(x)

def pinv(x):
	return np.linalg.pinv(x)

def arctan(x):
	return np.arctan(x)

def vstack(x):
	return np.vstack(x)

def abs(x):
	return np.abs(x)

def max(x):
	return np.max(x)

def ones(x):
	return np.ones(x)

def zeros(x):
	return np.zeros(x)

def empty(x):
	return np.empty(x)

def mean(x, axis=None):
	return np.mean(x, axis)

def sqrt(x):
	return np.sqrt(x)

def identity(x):
	return np.identity(x)

def rand(*x):
	return np.random.rand(*x)

def power(x, y):
	return np.power(x, y)

def exp(x):
	return np.exp(x)

def cosh(x):
    return np.cosh(x)  

def log(x):
    return np.log(x)

def tanh(x):
	return np.tanh(x)

def concatenate(tuple, axis=0):
	return np.concatenate(tuple, axis=axis)

def sign(x):
	return np.sign(x)

def argmax(x, axis):
    return np.argmax(x, axis)

def zeros_like(x):
    return np.zeros_like(x)

def all(x):
	return np.all(x)

def correlate(a, v, mode='valid'):
	return np.correlate(a, v, mode='valid')

def var(x):
	return np.var(x)

def allclose(x, y, atol=1e-05, rtol=0, equal_nan=False):
    return np.allclose(x, y, atol, atol, equal_nan)

def ptp(x, axis=None):
    return np.ptp(x, axis)


"""
    Implementation of the general ESN model.
"""

import numpy as np
import numpy.random as rnd
from .BaseESN import BaseESN

from . import backend as B

from sklearn.linear_model import Ridge
from sklearn.svm import SVR
from sklearn.linear_model import LogisticRegression
import progressbar

from .optimizers import GradientOptimizer
from .optimizers import GridSearchOptimizer


class PredictionESN(BaseESN):
    def __init__(self, n_input, n_reservoir, n_output,
                 spectralRadius=1.0, noiseLevel=0.0, inputScaling=None,
                 leakingRate=1.0, feedbackScaling = 1.0, reservoirDensity=0.2, randomSeed=None,
                 out_activation=lambda x: x, out_inverse_activation=lambda x: x,
                 weightGeneration='naive', bias=1.0, outputBias=1.0, feedback = False,
                 outputInputScaling=1.0, input_density=1.0, solver='pinv', regressionParameters={}, activation = B.tanh, activationDerivation=lambda x: 1.0/B.cosh(x)**2):

        super(PredictionESN, self).__init__(n_input=n_input, n_reservoir=n_reservoir, n_output=n_output, spectralRadius=spectralRadius,
                                  noiseLevel=noiseLevel, inputScaling=inputScaling, leakingRate=leakingRate, feedbackScaling = feedbackScaling, reservoirDensity=reservoirDensity,
                                  randomSeed=randomSeed, feedback = feedback, out_activation=out_activation, out_inverse_activation=out_inverse_activation,
                                  weightGeneration=weightGeneration, bias=bias, outputBias=outputBias, outputInputScaling=outputInputScaling,
                                  input_density=input_density, activation=activation, activationDerivation=activationDerivation)


        self._solver = solver
        self._regressionParameters = regressionParameters

        self._x = B.zeros((self.n_reservoir, 1))

        """
            allowed values for the solver:
                pinv
                lsqr (will only be used in the thesis)

                sklearn_auto
                sklearn_svd
                sklearn_cholesky
                sklearn_lsqr
                sklearn_sag
        """


    """
        Fits the ESN so that by applying the inputData the outputData will be produced.
    """
    def fit(self, inputData, outputData, transientTime="AutoReduce", transientTimeCalculationEpsilon = 1e-3, transientTimeCalculationLength = 20, verbose=0):
        #check the input data
        if self.n_input != 0:
            if len(inputData.shape) == 3 and len(outputData.shape) > 1:
                #multiple time series are used with a shape (timeseries, time, dimension) -> (timeseries, time, dimension)
                if inputData.shape[0] != outputData.shape[0]:
                    raise ValueError("Amount of input and output datasets is not equal - {0} != {1}".format(inputData.shape[0], outputData.shape[0]))
                if inputData.shape[1] != outputData.shape[1]:
                    raise ValueError("Amount of input and output time steps is not equal - {0} != {1}".format(inputData.shape[1], outputData.shape[1]))
            else:
                if inputData.shape[0] != outputData.shape[0]:
                    raise ValueError("Amount of input and output time steps is not equal - {0} != {1}".format(inputData.shape[0], outputData.shape[0]))
        else:
            if inputData is not None:
                raise ValueError("n_input has been set to zero. Therefore, the given inputData will not be used.")

        #reshape the input/output data to have the shape (timeseries, time, dimension)
        if len(outputData.shape) <= 2:
            outputData = outputData.reshape((1, -1, 1))
        if len(inputData.shape) <= 2:
            inputData = inputData.reshape((1, -1, 1))

        if inputData is not None:
            inputData = B.array(inputData)
        if outputData is not None:
            outputData = B.array(outputData)
        
        self.resetState()

        # Automatic transient time calculations
        if transientTime == "Auto":
            transientTime = self.calculateTransientTime(inputData[0], outputData[0], transientTimeCalculationEpsilon, transientTimeCalculationLength)
        if transientTime == "AutoReduce":
            if (inputData is None and outputData.shape[2] == 1) or inputData.shape[2] == 1:
                transientTime = self.calculateTransientTime(inputData[0], outputData[0], transientTimeCalculationEpsilon, transientTimeCalculationLength)
                transientTime = self.reduceTransientTime(inputData[0], outputData[0], transientTime)
            else:
                print("Transient time reduction is supported only for 1 dimensional input.")

        partialLength = (inputData.shape[1]-transientTime)
        totalLength = inputData.shape[0] * partialLength
        self._X = B.empty((1 + self.n_input + self.n_reservoir, totalLength))

        if (verbose > 0):
            bar = progressbar.ProgressBar(max_value=inputData.shape[0], redirect_stdout=True, poll_interval=0.0001)
            bar.update(0)

        for i in range(inputData.shape[0]):
            self._X[:, i*partialLength:(i+1)*partialLength] = self.propagate(inputData[i], outputData[i], transientTime, verbose-1)
            if (verbose > 0):
                bar.update(i)
        if (verbose > 0):
            bar.finish()


        #define the target values
        Y_target = B.empty((outputData.shape[2], totalLength))
        for i in range(inputData.shape[0]):
            Y_target[:, i*partialLength:(i+1)*partialLength] = self.out_inverse_activation(outputData[i]).T[:,transientTime:]

        if (self._solver == "pinv"):
            self._WOut = B.dot(Y_target, B.pinv(self._X))

            #calculate the training prediction now
            train_prediction = self.out_activation((B.dot(self._WOut, self._X)).T)

        elif (self._solver == "lsqr"):
            X_T = self._X.T
            self._WOut = B.dot(B.dot(Y_target, X_T),B.inv(B.dot(self._X,X_T) + self._regressionParameters[0]*B.identity(1+self.n_input+self.n_reservoir)))

            """
                #alternative represantation of the equation

                Xt = X.T

                A = np.dot(X, Y_target.T)

                B = np.linalg.inv(np.dot(X, Xt)  + regression_parameter*np.identity(1+self.n_input+self.n_reservoir))

                self._WOut = np.dot(B, A)
                self._WOut = self._WOut.T
            """

            #calculate the training prediction now
            train_prediction = self.out_activation(B.dot(self._WOut, self._X).T)

        elif (self._solver in ["sklearn_auto", "sklearn_lsqr", "sklearn_sag", "sklearn_svd"]):
            mode = self._solver[8:]
            params = self._regressionParameters
            params["solver"] = mode
            self._ridgeSolver = Ridge(**params)

            self._ridgeSolver.fit(self._X.T, Y_target.T)

            #calculate the training prediction now
            train_prediction = self.out_activation(self._ridgeSolver.predict(self._X.T))

        elif (self._solver in ["sklearn_svr", "sklearn_svc"]):
            self._ridgeSolver = SVR(**self._regressionParameters)

            self._ridgeSolver.fit(self._X.T, Y_target.T.ravel())

            #calculate the training prediction now
            train_prediction = self.out_activation(self._ridgeSolver.predict(self._X.T))

        #calculate the training error now
        #flatten the outputData
        outputData = outputData[:, transientTime:, :].reshape(totalLength, -1)
        training_error = B.sqrt(B.mean((train_prediction - outputData)**2))
        return training_error


    """
        Use the ESN in the generative mode to generate a signal autonomously.
    """
    def generate(self, n, inputData=None, continuation=True, initialData=None, update_processor=lambda x:x, verbose=0):
        #check the input data
        #if (self.n_input != self.n_output):
        #    raise ValueError("n_input does not equal n_output. The generation mode uses its own output as its input. Therefore, n_input has to be equal to n_output - please adjust these numbers!")

        if inputData is not None:
            inputData = B.array(inputData)
        if initialData is not None:
            initialData = B.array(initialData)

        #let some input run through the ESN to initialize its states from a new starting value
        if not continuation:
            self._x = B.zeros(self._x.shape)

            if initialData is not None:
                if type(initialData) is tuple:
                    initialDataInput, initialDataOutput = initialData 
                    if initialDataInput is not None and len(initialDataInput) != len(initialDataOutput):
                        raise ValueError("Length of the inputData and the outputData of the initialData tuple do not match.")
                else:
                    raise ValueError("initialData has to be a tuple consisting out of the input and the output data.")

                for t in range(initialDataInput.shape[0]):
                    super(PredictionESN, self).update(initialDataInput[t], initialDataOutput[t])

        if self.n_input != 0:
            if inputData is None:
                raise ValueError("inputData must not be None.")
            elif len(inputData) < n:
                raise ValueError("Length of inputData has to be >= n.")

        _, Y = self.propagate(inputData, None, verbose=verbose, steps=n)
        Y = update_processor(Y)
        
        #return the result
        return Y.T

    """
        Use the ESN in the predictive mode to predict the output signal by using an input signal.
    """
    def predict(self, inputData, continuation=True, initialData=None, update_processor=lambda x:x, verbose=0):
        inputData = B.array(inputData)
        
        #let some input run through the ESN to initialize its states from a new starting value
        if (not continuation):
            self._x = B.zeros(self._x.shape)

            if initialData is not None:
                if self._WFeedback is None:
                    for t in range(initialData.shape[0]):
                        super(PredictionESN, self).update(initialData[t])
                else:   
                    if type(initialData) is tuple:
                        initialDataInput, initialDataOutput = initialData 
                        if (len(initialDataInput) != len(initialDataOutput)):
                            raise ValueError("Length of the inputData and the outputData of the initialData tuple do not match.")
                    else:
                        raise ValueError("initialData has to be a tuple consisting out of the input and the output data.")

                    super(PredictionESN, self).update(initialDataInput[t], initialDataOutput[t])


        X = self.propagate(inputData, verbose=verbose)

        if self._WFeedback is not None:
            X, _ = X

        #calculate the prediction using the trained model
        if (self._solver in ["sklearn_auto", "sklearn_lsqr", "sklearn_sag", "sklearn_svd", "sklearn_svr"]):
            Y = self._ridgeSolver.predict(X.T).reshape((self.n_output, -1))
        else:
            Y = B.dot(self._WOut, X)

        #apply the output activation function
        Y = update_processor(self.out_activation(Y))
     
        #return the result
        return Y.T

    def optimize(self, trainingInput, trainingOutput, validationInput, validationOutput, verbose):
        gridSearch = GridSearchOptimizer()
        gradientOptimizer = GradientOptimizer()
        pipe = Pipeline(gridSearch, gradientOptimizer)

        pipe.fit(trainingInput, trainingOutput, validationInput, validationOutput, verbose)


from setuptools import setup, find_packages

setup(name='easyesn',
      version='0.1.3.4',
      description='',
      url='https://github.com/kalekiu/easyesn',
      author='Roland Zimmermann, Luca Thiede',
      author_email='support@flashtek.de',
      license='MIT',
      packages=find_packages(),
      install_requires=[
            'numpy',
            'progressbar2',
            'dill',
            'multiprocess'
      ],
      classifiers=[
            'Intended Audience :: Science/Research',
            'Intended Audience :: Developers',
            'Topic :: Scientific/Engineering',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'Programming Language :: Python :: 3'
      ],
      zip_safe=False)


from numpy import *
import matplotlib.pyplot as plt

class Reservoir:

    def __init__(self, input_dim, target_dim, size, spectralRadius=0.9, feedbackScaling=1, inputScaling=1, leakingRate=0.3,
                 randomState=42, density=0.1, transientTime=0):
        random.seed(randomState)
        self.input_dim = input_dim
        self.target_dim = target_dim
        self.size = size
        self.spectralRadius = spectralRadius
        self.inputScaling = inputScaling
        self.leakingRate = leakingRate
        self.density = density
        self.transientTime = transientTime

        self.generateReservoirConnections()



    def generateReservoirConnections(self):
        # generate input matrix
        self.W_in_uniform = (random.rand(self.size, self.input_dim + 1) - 0.5)
        self.W_in = self.inputScaling * self.W_in_uniform

        # generate dense random matrix
        self.W_rand = (random.rand(self.size, self.size) - 0.5)

        # generate sparse topology matrix with either 0 or 1
        self.W_top = random.rand(self.size, self.size)
        self.W_top[where(self.W_top < 1 - self.density)] = 0
        self.W_top[where(self.W_top >= 1 - self.density)] = 1

        # multiply both for reservoir matrix and scale reservoir to spectral radius
        print("Calculate spectral radius")
        self.setTopology(self.W_top)
        print("Done")


    def setTopology(self, W_top):
        self.W_top = W_top
        self.W_uniform = W_top * self.W_rand

        # scale spectral radius
        biggestEigenvalue = max(abs(linalg.eig(self.W_uniform)[0]))
        self.W_uniform = self.W_uniform / biggestEigenvalue
        self.W = self.W_uniform * self.spectralRadius

    # rescale spectral radius
    def tuneSpectralRadius(self, newSpectralRadius):
        self.W = self.W_uniform * newSpectralRadius
        self.spectralRadius = newSpectralRadius

    # rescale input scaling
    def tuneInputScaling(self, newInputScaling):
        self.W_in = self.W_in_uniform * newInputScaling
        self.inputScaling = newInputScaling

    # rescale leaking rate
    def tuneLeakingRate(self, newLeakingRate):
        self.leakingRate = newLeakingRate

    # calculate RMSE
    def Loss(self, prediction, target):
        return sqrt(((target - prediction) ** 2).mean(axis=None))

    def ridgeRegression(self, Y, X, reg):
        X_T = X.T
        return dot(dot(Y, X_T), linalg.inv(dot(X, X_T) + reg * eye(1 + self.input_dim + self.size)))

    # get reservoir output
    def readOut(self, u, x):
        return dot(self.W_out, vstack((1, u, x)))

    # terms within nonlinearity
    def X(self, x, u):
        return dot(self.W_in, vstack((1,u))) + dot(self.W, x)

    def f(self, x):
        return tanh(x)

    #update step of the reservoir states
    def updateNeuronState(self, x, u):
        newNeuronState = self.f(self.X(x, u))
        return (1 - self.leakingRate) * x + self.leakingRate * newNeuronState



################ Fitting and predicting with reservoir ################
    def fit(self, inputs, targets, trainLength, penalty=0.1, errorEvaluationLength=500):

        # initialize
        self.designMatrix = zeros((1 + self.input_dim + self.size, trainLength - self.transientTime))
        Ytarget = targets[None, self.transientTime:trainLength]
        self.x = zeros((self.size, 1))


        for t in range(trainLength):
            u = inputs[t].reshape(-1,1)
            self.x = self.updateNeuronState(self.x, u)
            if t >= self.transientTime:
                self.designMatrix[:, t - self.transientTime] = vstack((1, u, self.x))[:, 0]

        # compute W_out via ridge regression
        self.W_out = self.ridgeRegression(Ytarget, self.designMatrix, penalty)


        # calculate fit and evaluation loss
        self.prediction = dot(self.W_out, self.designMatrix)
        fitLoss = self.Loss(self.prediction, targets[self.transientTime :trainLength ])

        EvaPrediction = self.predictOnePointAhead(errorEvaluationLength, inputs, trainLength)
        evaLoss = self.Loss(EvaPrediction, targets[trainLength :trainLength + errorEvaluationLength])
        print("Evaluation RMSE= " + str(evaLoss) + " and Fit RMSE= " + str(fitLoss))

        return (evaLoss, fitLoss)



    def predictOnePointAhead(self, predictionLength, data, lastPoint):
        prediction = zeros((self.target_dim, predictionLength))
        x = copy(self.x)

        for t in range(predictionLength):
            u = data[lastPoint + t].reshape(-1,1)
            x = self.updateNeuronState(x, u)
            predictionPoint = dot(self.W_out, vstack((1, u, x)))
            prediction[:, t] = predictionPoint


        return prediction


################################ Optimization ###################################

    # f`(X)
    def activationDerivation(self, X):
        return 4 / (2 + exp(2 * X) + exp(-2 * X))

    # del x / del alpha
    def derivationForLeakingRate(self, oldDerivative, u, x):
        a = self.leakingRate
        X = self.X(x, u)
        return (1-a) * oldDerivative - x + self.f(X) + a * self.activationDerivation(X) * dot(self.W, oldDerivative )

    # del x / del rho
    def derivationForSpectralRadius(self, oldDerivative, u, x):
        a = self.leakingRate
        X = self.X(x, u)
        return (1-a) * oldDerivative + a * self.activationDerivation(X) * ( dot(self.W, oldDerivative) + dot(self.W_uniform, x) )

    # del x / del s_in
    def derivationForInputScaling(self, oldDerivative, u, x):
        a = self.leakingRate
        X = self.X(x, u)
        u = vstack((1,u))
        return (1-a)  * oldDerivative + a * self.activationDerivation(X) * ( dot(self.W, oldDerivative) + dot(self.W_in_uniform, u) )

    # del W_out / del beta
    def derivationForPenalty(self, Y, X, penalty):
        X_T = X.T
        term2 = linalg.inv( dot( X, X_T ) + penalty * eye(1 + self.input_dim + self.size) )
        return - dot( dot( Y, X_T ), dot( term2, term2 ) )

    # del W_out / del (alpha, rho or s_in)
    def derivationWoutForP(self, Y, X, XPrime, penalty):
        X_T = X.T
        XPrime_T = XPrime.T

        # A = dot(X,X_T) + penalty*eye(1 + self.target_dim + self.size)
        # APrime = dot( XPrime, X_T) + dot( X, XPrime_T )
        # APrime_T = APrime.T
        # InvA = linalg.inv(A)
        # InvA_T = InvA.T
        #
        # term1 = dot(XPrime_T, InvA)
        #
        # term21 = -dot( InvA, dot( APrime, InvA ) )
        # term22 = dot( dot( dot( InvA, InvA_T), APrime_T), eye(1 + self.target_dim + self.size) - dot( A, InvA ) )
        # term23 = dot( dot( eye(1 + self.target_dim + self.size) - dot( InvA, A ), APrime_T), dot( InvA_T, InvA) )
        # term2 = dot( X_T, term21 + term22 + term23 )
        #
        # return dot( Y, term1 + term2)

        term1 = linalg.inv(dot(X,X_T) + penalty*eye(1 + self.input_dim + self.size))
        term2 = dot( XPrime, X_T) + dot( X, XPrime_T )

        return dot( Y, dot( XPrime_T, term1 ) - dot( dot( dot( X_T, term1 ), term2 ), term1 ) )






    def optimizeParameterForTrainError(self, inputs, targets, trainLength, learningRate = 0.0001, epochs=1, penalty=0.1, errorEvaluationLength=500):
        # initializations of arrays:
        Ytarget = targets[None, self.transientTime:trainLength]

        # initializations for plotting parameter and losses at the end
        inputScalings = list()
        leakingRates = list()
        spectralRadiuses = list()
        fitLosses = list()
        evaLosses = list()

        # initializations for arrays which collect all the gradients of the error of the single time steps, which get add at the end
        srGradients = zeros(trainLength - self.transientTime)
        lrGradients = zeros(trainLength - self.transientTime)
        isGradients = zeros(trainLength - self.transientTime)

        # collecting the single derivatives  - > this is the derivation of design matrix when filled
        srGradientsMatrix = zeros((self.size + self.input_dim + 1, trainLength - self.transientTime))
        lrGradientsMatrix = zeros((self.size + self.input_dim + 1, trainLength - self.transientTime))
        isGradientsMatrix = zeros((self.size + self.input_dim + 1, trainLength - self.transientTime))

        # initialize fallback Parameter
        oldSR = self.spectralRadius
        oldLR = self.leakingRate
        oldIS = self.inputScaling

        # initialize self.designMatrix and self.W_out
        _, oldLoss, = self.fit(inputs, targets, trainLength, penalty=penalty, errorEvaluationLength=1)


        for epoch in range(epochs):
            print("###################### Start epoch: " + str(epoch) + " ##########################")
            # initialize del x / del (rho, alpha, s_in) and reservoir state itself
            derivationSpectralRadius = zeros((self.size, 1))
            derivationLeakingRate = zeros((self.size, 1))
            derivationInputScaling = zeros((self.size, 1))

            # initialize the neuron states new
            x = zeros((self.size, 1))

            # go thorugh the train length (e.g. the time, on which W_out gets calculated)
            for t in range(trainLength):
                u = inputs[t].reshape(-1, 1)
                oldx = x
                x = self.updateNeuronState(x, u)

                # calculate the del /x del (rho, alpha, s_in)
                derivationSpectralRadius = self.derivationForSpectralRadius(derivationSpectralRadius, u, oldx)
                derivationLeakingRate = self.derivationForLeakingRate(derivationLeakingRate, u, oldx)
                derivationInputScaling = self.derivationForInputScaling(derivationInputScaling, u, oldx)

                if t >= self.transientTime:

                    # concatenate with zeros (for the derivatives of the input and the 1, which are always 0)
                    derivationConcatinationSpectralRadius = concatenate((zeros(self.input_dim + 1), derivationSpectralRadius[:, 0]), axis=0)
                    derivationConcatinationLeakingRate = concatenate((zeros(self.input_dim + 1), derivationLeakingRate[:, 0]), axis=0)
                    derivationConcatinationInputScaling = concatenate((zeros(self.input_dim + 1), derivationInputScaling[:, 0]), axis=0)

                    # add to matrix
                    srGradientsMatrix[:, t - self.transientTime] = derivationConcatinationSpectralRadius
                    lrGradientsMatrix[:, t - self.transientTime] = derivationConcatinationLeakingRate
                    isGradientsMatrix[:, t - self.transientTime] = derivationConcatinationInputScaling

            # calculate del W_out / del (rho, alpha, s_in) based on the designMatrix and the derivative of the designMatrix we just calculated
            WoutPrimeSR = self.derivationWoutForP(Ytarget, self.designMatrix, srGradientsMatrix, penalty)
            WoutPrimeLR = self.derivationWoutForP(Ytarget, self.designMatrix, lrGradientsMatrix, penalty)
            WoutPrimeIS = self.derivationWoutForP(Ytarget, self.designMatrix, isGradientsMatrix, penalty)

            # reinitialize the states
            x = zeros((self.size, 1))
            # go through the train time again, and this time, calculate del error / del (rho, alpha, s_in) based on del W_out and the single derivatives
            for t in range(trainLength):
                u = inputs[t].reshape(-1, 1)
                x = self.updateNeuronState(x, u)

                if t >= self.transientTime:
                    # calculate error at given time step
                    error = (targets[t] - self.readOut(u, x)).T

                    # calculate gradients
                    gradientSR = dot(-error, dot( WoutPrimeSR, vstack((1, u, x))[:, 0] ) + dot(self.W_out, srGradientsMatrix[:,t-self.transientTime]) )
                    srGradients[t - self.transientTime] = gradientSR

                    gradientLR = dot(-error, dot( WoutPrimeLR, vstack((1, u, x))[:, 0] ) + dot(self.W_out, lrGradientsMatrix[:,t-self.transientTime]) )
                    lrGradients[t - self.transientTime] = gradientLR

                    gradientIS = dot(-error, dot( WoutPrimeIS, vstack((1, u, x))[:, 0] ) + dot(self.W_out, isGradientsMatrix[:,t-self.transientTime]) )
                    isGradients[t - self.transientTime] = gradientIS


            # sum up the gradients del error / del (rho, alpha, s_in) to get final gradient
            gradientSR = sum(srGradients)
            gradientLR = sum(lrGradients)
            gradientIS = sum(isGradients)


            # normalize gradients to length 1
            gradientVectorLength = sqrt(gradientSR ** 2 + gradientLR ** 2 + gradientIS ** 2)
            #gradientVectorLength = sqrt(gradientSR ** 2 + gradientLR ** 2)
            #gradientVectorLength = sqrt(gradientSR ** 2)

            gradientSR /= gradientVectorLength
            gradientLR /= gradientVectorLength
            gradientIS /= gradientVectorLength

            # update spectral radius
            self.tuneSpectralRadius(self.spectralRadius - learningRate * gradientSR)

            # update leaking rate
            self.leakingRate = self.leakingRate - learningRate * gradientLR

            # update input scaling
            self.tuneInputScaling(self.inputScaling - learningRate * gradientIS)

            # calculate the errors and update the self.designMatrix and the W_out
            evaLoss, fitLoss = self.fit(inputs, targets, trainLength, penalty=penalty, errorEvaluationLength=errorEvaluationLength)

            if fitLoss > oldLoss:
                self.tuneSpectralRadius(oldSR)
                self.leakingRate = oldLR
                self.tuneInputScaling(oldIS)
                learningRate = learningRate / 2
            else:
                oldSR = self.spectralRadius
                oldLR = self.leakingRate
                oldIS = self.inputScaling
                oldLoss = fitLoss

                spectralRadiuses.append(self.spectralRadius)
                leakingRates.append(self.leakingRate)
                inputScalings.append(self.inputScaling)

                fitLosses.append(fitLoss)
                evaLosses.append(evaLoss)


        evaLoss = evaLosses[-1]
        fitLoss = fitLosses[-1]

        return (evaLoss, fitLoss, evaLosses, fitLosses, inputScalings, leakingRates, spectralRadiuses)



####################################################################################################################################################


    def optimizeParameterForEvaluationError(self, inputs, targets,  trainLength, optimizationLength, learningRate = 0.0001, epochs=1, penalty=0.1):
        # initializations of arrays:
        Ytarget = targets[None, self.transientTime:trainLength]

        # initializations for plotting parameter and losses at the end
        inputScalings = list()
        leakingRates = list()
        spectralRadiuses = list()
        fitLosses = list()
        evaLosses = list()

        # initializations for arrays which collect all the gradients of the error of the single time steps, which get add at the end
        srGradients = zeros(optimizationLength)
        lrGradients = zeros(optimizationLength)
        isGradients = zeros(optimizationLength)

        # collecting the single derivatives  - > this is the derivation of design matrix when filled
        srGradientsMatrix = zeros((self.size + self.input_dim + 1, trainLength - self.transientTime))
        lrGradientsMatrix = zeros((self.size + self.input_dim + 1, trainLength - self.transientTime))
        isGradientsMatrix = zeros((self.size + self.input_dim + 1, trainLength - self.transientTime))

        # initialize variables for the "when the error goes up, go back and divide learning rate by 2" mechanism
        oldSR = self.spectralRadius
        oldLR = self.leakingRate
        oldIS = self.inputScaling

        # initialize self.designMatrix and self.W_out
        oldLoss, _, = self.fit(inputs, targets, trainLength, penalty=penalty, errorEvaluationLength=optimizationLength)

        for epoch in range(epochs):
            print("###################### Start epoch: " + str(epoch) + " ##########################")

            # initialize del x / del (rho, alpha, s_in) and reservoir state itself
            derivationSpectralRadius = zeros((self.size, 1))
            derivationLeakingRate = zeros((self.size, 1))
            derivationInputScaling = zeros((self.size, 1))
            x = zeros((self.size, 1))

            # go thorugh the train length (e.g. the time, on which W_out gets calculated)
            for t in range(trainLength):
                u = inputs[t].reshape(-1, 1)
                oldx = x
                x = self.updateNeuronState(x, u)

                # calculate the del x/ del (rho, alpha, s_in)
                derivationSpectralRadius = self.derivationForSpectralRadius(derivationSpectralRadius, u, oldx)
                derivationLeakingRate = self.derivationForLeakingRate(derivationLeakingRate, u, oldx)
                derivationInputScaling = self.derivationForInputScaling(derivationInputScaling, u, oldx)

                if t >= self.transientTime:

                    # concatenate with zeros (for the derivatives of the input and the 1, which are always 0)
                    srGradientsMatrix[:, t - self.transientTime] = concatenate((zeros(self.input_dim + 1), derivationSpectralRadius[:, 0]), axis=0)
                    lrGradientsMatrix[:, t - self.transientTime] = concatenate((zeros(self.input_dim + 1), derivationLeakingRate[:, 0]), axis=0)
                    isGradientsMatrix[:, t - self.transientTime] = concatenate((zeros(self.input_dim + 1), derivationInputScaling[:, 0]), axis=0)


            # add to matrix
            WoutPrimeSR = self.derivationWoutForP(Ytarget, self.designMatrix, srGradientsMatrix, penalty)
            WoutPrimeLR = self.derivationWoutForP(Ytarget, self.designMatrix, lrGradientsMatrix, penalty)
            WoutPrimeIS = self.derivationWoutForP(Ytarget, self.designMatrix, isGradientsMatrix, penalty)

            # this time go through validation length
            for t in range(optimizationLength):
                u = inputs[t + trainLength].reshape(-1, 1)
                oldx = x
                x = self.updateNeuronState(x, u)

                # calculate error at given time step
                error = ( targets[t + trainLength] - self.readOut(u, x) ).T

                # calculate del x / del (rho, alpha, s_in)
                derivationSpectralRadius = self.derivationForSpectralRadius(derivationSpectralRadius, u, oldx)
                derivationLeakingRate = self.derivationForLeakingRate(derivationLeakingRate, u, oldx)
                derivationInputScaling = self.derivationForInputScaling(derivationInputScaling, u, oldx)

                # concatenate derivations with 0
                derivationConcatinationSpectralRadius = concatenate((zeros(self.input_dim + 1), derivationSpectralRadius[:, 0]), axis=0)
                derivationConcatinationLeakingRate = concatenate((zeros(self.input_dim + 1), derivationLeakingRate[:, 0]), axis=0)
                derivationConcatinationInputScaling = concatenate((zeros(self.input_dim + 1), derivationInputScaling[:, 0]), axis=0)

                # calculate gradients
                gradientSR = dot(-error, dot(self.W_out, derivationConcatinationSpectralRadius) + dot( WoutPrimeSR, vstack((1, u, x))[:, 0] ) )
                srGradients[t] = gradientSR

                gradientLR = dot(-error, dot(self.W_out, derivationConcatinationLeakingRate) + dot( WoutPrimeLR, vstack((1, u, x))[:, 0] ) )
                lrGradients[t] = gradientLR

                gradientIS = dot(-error, dot(self.W_out, derivationConcatinationInputScaling) + dot( WoutPrimeIS, vstack((1, u, x))[:, 0] ) )
                isGradients[t] = gradientIS


            # sum up the gradients del error / del (rho, alpha, s_in) to get final gradient
            gradientSR = sum(srGradients)
            gradientLR = sum(lrGradients)
            gradientIS = sum(isGradients)

            # normalize length of gradient to 1
            gradientVectorLength = sqrt(gradientSR ** 2 + gradientLR ** 2 + gradientIS ** 2)
            #gradientVectorLength = sqrt(gradientIS ** 2 + gradientLR ** 2 )

            gradientSR /= gradientVectorLength
            gradientLR /= gradientVectorLength
            gradientIS /= gradientVectorLength

            # update spectral radius
            self.tuneSpectralRadius(self.spectralRadius - learningRate * gradientSR)

            # update leaking rate
            self.leakingRate = self.leakingRate - learningRate * gradientLR

            # update input scaling
            self.tuneInputScaling(self.inputScaling - learningRate * gradientIS)


            # calculate the errors and update the self.designMatrix and the W_out
            evaLoss, fitLoss = self.fit(inputs, targets, trainLength, penalty=penalty, errorEvaluationLength=optimizationLength)

            # this is the "when the error goes up, go back and divide learning rate by 2" mechanism
            if evaLoss > oldLoss:
                self.tuneSpectralRadius(oldSR)
                self.leakingRate = oldLR
                self.tuneInputScaling(oldIS)
                learningRate = learningRate / 2
            else:
                oldSR = self.spectralRadius
                oldLR = self.leakingRate
                oldIS = self.inputScaling
                oldLoss = evaLoss

                spectralRadiuses.append(self.spectralRadius)
                leakingRates.append(self.leakingRate)
                inputScalings.append(self.inputScaling)

                fitLosses.append(fitLoss)
                evaLosses.append(evaLoss)



        evaLoss = evaLosses[-1]
        fitLoss = fitLosses[-1]

        return (evaLoss, fitLoss, evaLosses, fitLosses, inputScalings, leakingRates, spectralRadiuses)


    def optimizePenaltyForEvaluationError(self, inputs, targets, trainLength, optimizationLength, learningRate = 0.0001, epochs=1, penalty=0.1):

        Ytarget = targets[None, self.transientTime:trainLength]
        penalty = penalty

        fitLosses = list()
        evaLosses = list()
        penalties = list()

        penaltyDerivatives = zeros(optimizationLength)

        oldPenalty = penalty

        oldLoss, fitLoss = self.fit(inputs, targets, trainLength, penalty=penalty, errorEvaluationLength=optimizationLength)

        evaluationEchoFunction = zeros((1+self.size+self.input_dim, optimizationLength))
        u = inputs[trainLength]
        x = self.x
        for t in range(optimizationLength):
            x = self.updateNeuronState(x, u)
            u = inputs[trainLength + t].reshape(-1, 1)
            evaluationEchoFunction[:,t] = vstack((1, u, x)).squeeze()
            # u = predictionPoint

        for epoch in range(epochs):
            print("###################### Start epoch: " + str(epoch) + " ##########################")

            penaltyDerivative = self.derivationForPenalty(Ytarget, self.designMatrix, penalty)

            for t in range(optimizationLength):
                predictionPoint = dot(self.W_out, evaluationEchoFunction[:,t].reshape(-1,1))
                error = ( targets[trainLength + t] - predictionPoint ).T
                penaltyDerivatives[t] = - dot( dot(error, penaltyDerivative), vstack((1, u, x))[:, 0] )
                u = inputs[trainLength + t].reshape(-1, 1)
                #u = predictionPoint

            penaltyGradient = sum(penaltyDerivatives)
            penaltyGradient /= sqrt(penaltyGradient**2)

            penalty = penalty - learningRate * penaltyGradient

            evaLoss, fitLoss = self.fit(inputs, targets, trainLength, penalty=penalty, errorEvaluationLength=optimizationLength)


            # this is the "when the error goes up, go back and divide learning rate by 2" mechanism
            if evaLoss > oldLoss:
                penalty = oldPenalty
                learningRate = learningRate / 2
            else:
                oldPenalty = penalty
                oldLoss = evaLoss

                fitLosses.append( fitLoss )
                evaLosses.append( evaLoss )
                penalties.append( penalty )

        return (evaLoss, fitLoss, evaLosses, fitLosses, penalties)



    def optimizeAllParameter(self, inputs, targets,  trainLength, optimizationLength, learningRate = 0.01, learningRatePenalty=0.0001, epochs=1, penalty=0.1):
        # initializations of arrays:
        Ytarget = targets[None, self.transientTime:trainLength]

        # initializations for plotting parameter and losses at the end
        inputScalings = list()
        leakingRates = list()
        spectralRadiuses = list()
        penalties = list()
        fitLosses = list()
        evaLosses = list()

        # initializations for arrays which collect all the gradients of the error of the single time steps, which get add at the end
        srGradients = zeros(optimizationLength)
        lrGradients = zeros(optimizationLength)
        isGradients = zeros(optimizationLength)
        penaltyDerivatives = zeros(optimizationLength)

        # collecting the single derivatives  - > this is the derivation of design matrix when filled
        srGradientsMatrix = zeros((self.size + self.input_dim + 1, trainLength - self.transientTime))
        lrGradientsMatrix = zeros((self.size + self.input_dim + 1, trainLength - self.transientTime))
        isGradientsMatrix = zeros((self.size + self.input_dim + 1, trainLength - self.transientTime))

        # initialize variables for the "when the error goes up, go back and divide learning rate by 2" mechanism
        oldSR = self.spectralRadius
        oldLR = self.leakingRate
        oldIS = self.inputScaling
        oldPenalty = penalty

        # initialize self.designMatrix and self.W_out
        oldLoss, _ = self.fit(inputs, targets, trainLength, penalty=penalty, errorEvaluationLength=optimizationLength)

        for epoch in range(epochs):
            print("###################### Start epoch: " + str(epoch) + " ##########################")

            # initialize del x / del (rho, alpha, s_in) and reservoir state itself
            derivationSpectralRadius = zeros((self.size, 1))
            derivationLeakingRate = zeros((self.size, 1))
            derivationInputScaling = zeros((self.size, 1))
            x = zeros((self.size, 1))

            # go thorugh the train length (e.g. the time, on which W_out gets calculated)
            for t in range(trainLength):
                u = inputs[t].reshape(-1, 1)
                oldx = x
                x = self.updateNeuronState(x, u)

                # calculate the del x/ del (rho, alpha, s_in)
                derivationSpectralRadius = self.derivationForSpectralRadius(derivationSpectralRadius, u, oldx)
                derivationLeakingRate = self.derivationForLeakingRate(derivationLeakingRate, u, oldx)
                derivationInputScaling = self.derivationForInputScaling(derivationInputScaling, u, oldx)

                if t >= self.transientTime:

                    # concatenate with zeros (for the derivatives of the input and the 1, which are always 0)
                    srGradientsMatrix[:, t - self.transientTime] = concatenate((zeros(self.input_dim + 1), derivationSpectralRadius[:, 0]), axis=0)
                    lrGradientsMatrix[:, t - self.transientTime] = concatenate((zeros(self.input_dim + 1), derivationLeakingRate[:, 0]), axis=0)
                    isGradientsMatrix[:, t - self.transientTime] = concatenate((zeros(self.input_dim + 1), derivationInputScaling[:, 0]), axis=0)


            # add to matrix
            WoutPrimeSR = self.derivationWoutForP(Ytarget, self.designMatrix, srGradientsMatrix, penalty)
            WoutPrimeLR = self.derivationWoutForP(Ytarget, self.designMatrix, lrGradientsMatrix, penalty)
            WoutPrimeIS = self.derivationWoutForP(Ytarget, self.designMatrix, isGradientsMatrix, penalty)

            # penaltyDerivation
            penaltyDerivative = self.derivationForPenalty(Ytarget, self.designMatrix, penalty)

            # this time go through validation length
            for t in range(optimizationLength):
                u = inputs[t + trainLength].reshape(-1, 1)
                oldx = x
                x = self.updateNeuronState(x, u)

                # calculate error at given time step
                error = ( targets[t + trainLength] - self.readOut(u, x) ).T

                # calculate del x / del (rho, alpha, s_in)
                derivationSpectralRadius = self.derivationForSpectralRadius(derivationSpectralRadius, u, oldx)
                derivationLeakingRate = self.derivationForLeakingRate(derivationLeakingRate, u, oldx)
                derivationInputScaling = self.derivationForInputScaling(derivationInputScaling, u, oldx)

                # concatenate derivations with 0
                derivationConcatinationSpectralRadius = concatenate((zeros(self.input_dim + 1), derivationSpectralRadius[:, 0]), axis=0)
                derivationConcatinationLeakingRate = concatenate((zeros(self.input_dim + 1), derivationLeakingRate[:, 0]), axis=0)
                derivationConcatinationInputScaling = concatenate((zeros(self.input_dim + 1), derivationInputScaling[:, 0]), axis=0)

                # calculate gradients
                gradientSR = dot(-error, dot(self.W_out, derivationConcatinationSpectralRadius) + dot( WoutPrimeSR, vstack((1, u, x))[:, 0] ) )
                srGradients[t] = gradientSR

                gradientLR = dot(-error, dot(self.W_out, derivationConcatinationLeakingRate) + dot( WoutPrimeLR, vstack((1, u, x))[:, 0] ) )
                lrGradients[t] = gradientLR

                gradientIS = dot(-error, dot(self.W_out, derivationConcatinationInputScaling) + dot( WoutPrimeIS, vstack((1, u, x))[:, 0] ) )
                isGradients[t] = gradientIS

                penaltyDerivatives[t] = - dot(dot(error, penaltyDerivative), vstack((1, u, x))[:, 0])


            # sum up the gradients del error / del (rho, alpha, s_in) to get final gradient
            gradientSR = sum(srGradients)
            gradientLR = sum(lrGradients)
            gradientIS = sum(isGradients)
            penaltyGradient = sum(penaltyDerivatives)

            # normalize length of gradient to 1
            gradientVectorLength = sqrt(gradientSR ** 2 + gradientLR ** 2 + gradientIS ** 2)
            penaltyGradientVectorLength = sqrt(penaltyGradient**2)

            gradientSR /= gradientVectorLength
            gradientLR /= gradientVectorLength
            gradientIS /= gradientVectorLength
            penaltyGradient /= penaltyGradientVectorLength


            # update spectral radius
            self.tuneSpectralRadius(self.spectralRadius - learningRate * gradientSR)

            # update leaking rate
            self.leakingRate = self.leakingRate - learningRate * gradientLR

            # update input scaling
            self.tuneInputScaling(self.inputScaling - learningRate * gradientIS)

            # update penalty
            penalty = penalty - learningRatePenalty * penaltyGradient


            # calculate the errors and update the self.designMatrix and the W_out
            evaLoss, fitLoss = self.fit(inputs, targets, trainLength, penalty=penalty, errorEvaluationLength=optimizationLength)

            # this is the "when the error goes up, go back and divide learning rate by 2" mechanism
            if evaLoss > oldLoss:
                self.tuneSpectralRadius(oldSR)
                self.leakingRate = oldLR
                self.tuneInputScaling(oldIS)
                penalty = oldPenalty
                learningRate = learningRate / 2
                learningRatePenalty = learningRatePenalty / 6
            else:
                oldSR = self.spectralRadius
                oldLR = self.leakingRate
                oldIS = self.inputScaling
                oldPenalty = penalty
                oldLoss = evaLoss

                spectralRadiuses.append(self.spectralRadius)
                leakingRates.append(self.leakingRate)
                inputScalings.append(self.inputScaling)
                penalties.append(penalty)

                fitLosses.append(fitLoss)
                evaLosses.append(evaLoss)



        evaLoss = evaLosses[-1]
        fitLoss = fitLosses[-1]

        return (evaLoss, fitLoss, evaLosses, fitLosses, inputScalings, leakingRates, spectralRadiuses, penalties)




# ### example usage ###
# import numpy as np
# def generateData(T):
#     x = np.arange(0,T*0.1,0.1)
#     return np.sin(x) + np.sin(0.51*x) +np.sin(0.22*x) + np.sin(0.1002*x) + np.sin(0.05343*x)#np.sin(x)
#
# data = generateData(10000)
#
# data = data/np.max(data)
# targetData = np.roll(data,-1)
#
# plt.plot(data)
# plt.xlim(0,3000)
# plt.xlabel("t")
# plt.ylabel("f(t)")
# plt.show()
#
#
# transientTime = 1500
# trainTime = 2500
# evaTime = 200
# epocsh = 60
#
# reservoir = Reservoir(1,1,20, density=0.2, spectralRadius=0.5, leakingRate=0.5, inputScaling=1, transientTime=transientTime)
#
# print("#################################################")
# evaLoss, fitLoss, evaLosses, fitLosses, inputScalings, leakingRates, spectralRadiuses, penalties = reservoir.optimizeParameterForTrainError(data, targetData, trainTime, epochs = epocsh, learningRate = 0.015, errorEvaluationLength=evaTime, penalty=0.005)
# #evaLoss, fitLoss, evaLosses, fitLosses, inputScalings, leakingRates, spectralRadiuses, penalties = reservoir.optimizeParameterForEvaluationError(data, targetData, trainTime, evaTime, epochs = epocsh, learningRate = 0.02, penalty=0.005)
# print("##################################################")
#
#
#
# ### plot results ###
# plt.plot(inputScalings)
# plt.title("InputScaling")
# plt.show()
# plt.plot(leakingRates)
# plt.title("LeakingRate")
# plt.show()
# plt.plot(spectralRadiuses)
# plt.title("SpectralRadius")
# plt.show()
# plt.plot(fitLosses)
# plt.title("Fit error")
# plt.show()
# plt.plot(evaLosses)
# plt.title("Evaluation Error")
# plt.show()
# plt.plot(penalties)
# plt.title("Penalties")
# plt.show()
#
#
# plt.plot(np.dot(reservoir.W_out, reservoir.designMatrix)[0])
# plt.title("Fit")
# plt.show()
# plt.plot(reservoir.designMatrix.T)
# plt.title("Echo functions")
# plt.show()
# plt.plot(targetData[trainTime:trainTime + evaTime], label="data")
# plt.plot(reservoir.predictOnePointAhead(evaTime, data, trainTime)[0], label="prediction")
# plt.legend()
# plt.show()
